from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import WorkCapacityStandardsInputs
from pythermalcomfort.classes_return import WorkCapacity


def work_capacity_niosh(
    wbgt: float | list[float],
    met: float | list[float],
) -> WorkCapacity:
    """Estimate work capacity due to heat based on NIOSH standards as described by Brode et al.

    Estimates the amount of work that will be done at a given WBGT and
    intensity of work as a percent. 100% means work is unaffected by heat. 0%
    means no work is done.

    The function definitions / parameters can be found in: 1. Bröde P, Fiala D,
    Lemke B, Kjellstrom T. Estimated work ability in warm outdoor environments
    depends on the chosen heat stress assessment metric. International Journal
    of Biometeorology. 2018 Mar;62(3):331 45.

    For a comparison of different functions see Fig 1 of Day E, Fankhauser S, Kingsmill
    N, Costa H, Mavrogianni A. Upholding labour productivity under climate
    change: an assessment of adaptation options. Climate Policy. 2019
    Mar;19(3):367 85.

    Parameters
    ----------
    wbgt : float or list of floats
        Wet bulb globe temperature, [°C].
    met : float or list of floats
        Metabolic heat production in Watts

    Returns
    -------
    WorkCapacity
        A dataclass containing the work capacity. See
        :py:class:`~pythermalcomfort.classes_return.WorkCapacity` for more details. To access the
        `capacity` value, use the `capacity` attribute of the returned `WorkCapacity` instance, e.g.,
        `result.capacity`.


    """
    # Validate inputs
    WorkCapacityStandardsInputs(wbgt=wbgt, met=met)

    wbgt = np.array(wbgt)
    met = np.array(met)

    met_rest = 117  # assumed resting metabolic rate

    wbgt_lim = 56.7 - 11.5 * np.log10(met)
    wbgt_lim_rest = 56.7 - 11.5 * np.log10(met_rest)
    capacity = ((wbgt_lim_rest - wbgt) / (wbgt_lim_rest - wbgt_lim)) * 100
    capacity = np.clip(capacity, 0, 100)

    return WorkCapacity(capacity=capacity)



import numpy as np

from pythermalcomfort.models.vertical_tmp_grad_ppd import vertical_tmp_grad_ppd
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_vertical_tmp_grad_ppd(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the output correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.VERTICAL_TMP_GRAD_PPD.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = vertical_tmp_grad_ppd(**inputs)

        validate_result(result, outputs, tolerance)

    # Test for ValueError
    np.isclose(
        vertical_tmp_grad_ppd(25, 25, 0.3, 50, 1.2, 0.5, 7).ppd_vg,
        np.nan,
        equal_nan=True,
    )



import os
import sys

sys.path.insert(0, os.path.abspath(".."))

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_rtd_theme",
]
source_suffix = ".rst"
master_doc = "index"
project = "pythermalcomfort"
year = "2025"
author = "Federico Tartarini"
project_copyright = f"{year}, {author}"
version = release = "3.4.4"

autodoc_typehints = "none"
pygments_style = "trac"
templates_path = ["."]
extlinks = {
    "issue": (
        "https://github.com/CenterForTheBuiltEnvironment/pythermalcomfort/issues/%s",
        "issue %s",
    ),
    "pr": (
        "https://github.com/CenterForTheBuiltEnvironment/pythermalcomfort/pull/%s",
        "PR %s",
    ),
}
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

html_theme = "pydata_sphinx_theme"

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False

html_title = f"pythermalcomfort {version}"
html_short_title = f"{project}-{version}"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False

html_sidebars = {
    "**": ["sidebar-nav-bs"],
    "installation": [],
    "contributing": [],
    "authors": [],
}

html_theme_options = {
    "icon_links": [
        {
            # Label for this link
            "name": "GitHub",
            # URL where the link will redirect
            "url": "https://github.com/CenterForTheBuiltEnvironment/pythermalcomfort",  # required
            # Icon class (if "type": "fontawesome"), or path to local image (if "type": "local")
            "icon": "fa-brands fa-square-github",
            # The type of image to be used (see below for details)
            "type": "fontawesome",
        },
        {
            "name": "LinkedIn",
            "url": "https://www.linkedin.com/in/federico-tartarini",  # required
            "icon": "fa-brands fa-linkedin",
            "type": "fontawesome",
        },
        {
            "name": "Google Scholar",
            "url": "https://scholar.google.com/citations?view_op=list_works&hl=en&hl=en&user=QcamSPwAAAAJ",
            "icon": "fa-brands fa-google-scholar",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/pythermalcomfort/",
            "icon": "fa-brands fa-python",
        },
    ],
    "secondary_sidebar_items": ["page-toc", "edit-this-page"],
    "content_footer_items": ["last-updated"],
}



import os
import time

import numpy as np
import pandas as pd

from pythermalcomfort.models import pmv_ppd_ashrae, pmv_ppd_iso
from pythermalcomfort.utilities import (
    clo_dynamic_ashrae,
    clo_individual_garments,
    met_typical_tasks,
    v_relative,
)

# input variables
tdb = 27  # dry bulb air temperature, [$^{\circ}$C]
tr = 25  # mean radiant temperature, [$^{\circ}$C]
v = 0.3  # average air speed, [m/s]
rh = 50  # relative humidity, [%]
activity = "Typing"  # participant's activity description
garments = ["Sweatpants", "T-shirt", "Shoes or sandals"]

met = met_typical_tasks[activity]  # activity met, [met]
icl = sum(
    [clo_individual_garments[item] for item in garments],
)  # calculate total clothing insulation

# calculate the relative air velocity
vr = v_relative(v=v, met=met)
# calculate the dynamic clothing insulation
clo = clo_dynamic_ashrae(clo=icl, met=met)

# calculate PMV in accordance with the ASHRAE 55 2020
results = pmv_ppd_iso(tdb=tdb, tr=tr, vr=vr, rh=rh, met=met, clo=clo)

# print the results
print(results)

# print PMV value
print(f"pmv={results['pmv']}, ppd={results['ppd']}%")

# for users who want to use the IP system
results_ip = pmv_ppd_iso(tdb=77, tr=77, vr=0.6, rh=50, met=1.1, clo=0.5, units="IP")
print(results_ip)

# If you want you can also pass pandas series or arrays as inputs
df = pd.read_csv(os.getcwd() + "/examples/template-SI.csv")

v_rel = v_relative(df["v"], df["met"])
clo_d = clo_dynamic_ashrae(df["clo"], df["met"])
results = pmv_ppd_ashrae(df["tdb"], df["tr"], v_rel, df["rh"], df["met"], clo_d, 0)
print(results)

df["vr"] = v_rel
df["clo_d"] = clo_d
df["pmv"] = results.pmv  # you can also use results["pmv"]
df["ppd"] = results["ppd"]

print(df.head())

# uncomment the following line if you want to save the data to .csv file
# df.to_csv('results.csv')

# This method is extremely fast and can perform a lot of calculations in very little time
iterations = 10000
tdb = np.empty(iterations)
tdb.fill(25)
tdb = tdb.tolist()
met = np.empty(iterations)
met.fill(1.5)
met = met.tolist()

v_rel = v_relative(0.1, met)
clo_d = clo_dynamic_ashrae(1, met)

# ASHRAE PMV
start = time.time()
pmv_ppd_ashrae(
    tdb=tdb,
    tr=23,
    vr=v_rel,
    rh=40,
    met=1.2,
    clo=clo_d,
)
end = time.time()
print(end - start)

# ISO PMV
start = time.time()
pmv_ppd_iso(tdb=tdb, tr=23, vr=v_rel, rh=40, met=1.2, clo=clo_d)
end = time.time()
print(end - start)



"""This code defines a set of models and constants to model heat exchange and
blood flow in different body parts and layers.
"""

import numpy as np

from pythermalcomfort.classes_return import JOS3BodyParts
from pythermalcomfort.jos3_functions.parameters import Default

LAYER_NAMES = ["artery", "vein", "sfvein", "core", "muscle", "fat", "skin"]


def index_order():
    """Define the index's order of the matrix.

    Returns
    -------
    index_dict : nested dictionary
        keys are the body names and layer names

    """
    # Defines existing layers as 1 or None
    index_dict = {}

    for key in ["head", "pelvis"]:
        index_dict[key] = {
            "artery": 1,
            "vein": 1,
            "sfvein": None,
            "core": 1,
            "muscle": 1,
            "fat": 1,
            "skin": 1,
        }

    for key in ["neck", "chest", "back"]:
        index_dict[key] = {
            "artery": 1,
            "vein": 1,
            "sfvein": None,
            "core": 1,
            "muscle": None,
            "fat": None,
            "skin": 1,
        }

    for key in JOS3BodyParts.get_attribute_names()[5:]:  # limb segments
        index_dict[key] = {
            "artery": 1,
            "vein": 1,
            "sfvein": 1,
            "core": 1,
            "muscle": None,
            "fat": None,
            "skin": 1,
        }

    # Sets ordered indices in the matrix
    index_dict["CB"] = 0
    order_count = 1
    for bn in JOS3BodyParts.get_attribute_names():
        for ln in LAYER_NAMES:
            if index_dict[bn][ln] is not None:
                index_dict[bn][ln] = order_count
                order_count += 1

    return index_dict, order_count


IDICT, NUM_NODES = index_order()


def index_by_layer(layer):
    """Get indices of the matrix by the layer name.

    Parameters
    ----------
    layer : str
        Layer name of jos.
        ex) artery, vein, sfvein, core, muscle, fat or skin.

    Returns
    -------
    indices of the matrix : list

    """
    # Gets indices by the layer name
    out_index = []
    for bn in JOS3BodyParts.get_attribute_names():
        for ln in LAYER_NAMES:
            if (layer.lower() == ln) and IDICT[bn][ln]:
                out_index.append(IDICT[bn][ln])
    return out_index


def valid_index_by_layer(layer):
    """Get indices of the matrix by the layer name.

    Parameters
    ----------
    layer : str
        Layer name of jos.
        ex) artery, vein, sfvein, core, muscle, fat or skin.

    Returns
    -------
    indices of the matrix : list

    """
    # Gets valid indices of the layer name
    out_index = []
    for i, bn in enumerate(JOS3BodyParts.get_attribute_names()):
        if IDICT[bn][layer]:
            out_index.append(i)
    return out_index


# Constant parameters of the matrix' indicies
INDEX = {}
VINDEX = {}
for key in LAYER_NAMES:
    INDEX[key] = index_by_layer(key)
    VINDEX[key] = valid_index_by_layer(key)


def local_arr(bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot):
    """Create matrix to calculate heat exchange by blood flow in each segment.

    [W/K]

    1.067 [Wh/(L･K)] * Bloodflow [L/h] = [W/K]
    """
    bf_local = np.zeros((NUM_NODES, NUM_NODES))
    for i, bn in enumerate(JOS3BodyParts.get_attribute_names()):
        # Dictionary of indecies in each body segment
        # key = layer name, value = index of matrix
        index_of = IDICT[bn]

        # Common
        bf_local[index_of["core"], index_of["artery"]] = 1.067 * bf_core[i]  # art to cr
        bf_local[index_of["skin"], index_of["artery"]] = 1.067 * bf_skin[i]  # art to sk
        bf_local[index_of["vein"], index_of["core"]] = 1.067 * bf_core[i]  # vein to cr
        bf_local[index_of["vein"], index_of["skin"]] = 1.067 * bf_skin[i]  # vein to sk

        # If the segment has a muslce or fat layer
        if index_of["muscle"] is not None:
            bf_local[index_of["muscle"], index_of["artery"]] = (
                1.067 * bf_muscle[i]
            )  # art to ms
            bf_local[index_of["vein"], index_of["muscle"]] = (
                1.067 * bf_muscle[i]
            )  # vein to ms
        if index_of["fat"] is not None:
            bf_local[index_of["fat"], index_of["artery"]] = (
                1.067 * bf_fat[i]
            )  # art to fat
            bf_local[index_of["vein"], index_of["fat"]] = (
                1.067 * bf_fat[i]
            )  # vein to fat

        # Only hand
        if i == 7 or i == 10:
            bf_local[index_of["sfvein"], index_of["artery"]] = (
                1.067 * bf_ava_hand
            )  # art to sfvein
        # Only foot
        if i == 13 or i == 16:
            bf_local[index_of["sfvein"], index_of["artery"]] = (
                1.067 * bf_ava_foot
            )  # art to sfvein

    return bf_local


def vessel_blood_flow(bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot):
    """Get artery and vein blood flow rate [l/h]."""
    xbf = bf_core + bf_muscle + bf_fat + bf_skin

    bf_art = np.zeros(Default.num_body_parts)
    bf_vein = np.zeros(Default.num_body_parts)

    # head
    bf_art[0] = xbf[0]
    bf_vein[0] = xbf[0]

    # neck (+head)
    bf_art[1] = xbf[1] + xbf[0]
    bf_vein[1] = xbf[1] + xbf[0]

    # chest
    bf_art[2] = xbf[2]
    bf_vein[2] = xbf[2]

    # back
    bf_art[3] = xbf[3]
    bf_vein[3] = xbf[3]

    # pelvis (+Thighs, Legs, Feet, AVA_Feet)
    bf_art[4] = xbf[4] + xbf[11 : Default.num_body_parts].sum() + 2 * bf_ava_foot
    bf_vein[4] = xbf[4] + xbf[11 : Default.num_body_parts].sum() + 2 * bf_ava_foot

    # L.Shoulder (+Arm, Hand, (arteryのみAVA_Hand))
    bf_art[5] = xbf[5:8].sum() + bf_ava_hand
    bf_vein[5] = xbf[5:8].sum()

    # L.Arm (+Hand)
    bf_art[6] = xbf[6:8].sum() + bf_ava_hand
    bf_vein[6] = xbf[6:8].sum()

    # L.Hand
    bf_art[7] = xbf[7] + bf_ava_hand
    bf_vein[7] = xbf[7]

    # R.Shoulder (+Arm, Hand, (arteryのみAVA_Hand))
    bf_art[8] = xbf[8:11].sum() + bf_ava_hand
    bf_vein[8] = xbf[8:11].sum()

    # R.Arm (+Hand)
    bf_art[9] = xbf[9:11].sum() + bf_ava_hand
    bf_vein[9] = xbf[9:11].sum()

    # R.Hand
    bf_art[10] = xbf[10] + bf_ava_hand
    bf_vein[10] = xbf[10]

    # L.Thigh (+Leg, Foot, (arteryのみAVA_Foot))
    bf_art[11] = xbf[11:14].sum() + bf_ava_foot
    bf_vein[11] = xbf[11:14].sum()

    # L.Leg (+Foot)
    bf_art[12] = xbf[12:14].sum() + bf_ava_foot
    bf_vein[12] = xbf[12:14].sum()

    # L.Foot
    bf_art[13] = xbf[13] + bf_ava_foot
    bf_vein[13] = xbf[13]

    # R.Thigh (+Leg, Foot, (arteryのみAVA_Foot))
    bf_art[14] = xbf[14 : Default.num_body_parts].sum() + bf_ava_foot
    bf_vein[14] = xbf[14 : Default.num_body_parts].sum()

    # R.Leg (+Foot)
    bf_art[15] = xbf[15 : Default.num_body_parts].sum() + bf_ava_foot
    bf_vein[15] = xbf[15 : Default.num_body_parts].sum()

    # R.Foot
    bf_art[16] = xbf[16] + bf_ava_foot
    bf_vein[16] = xbf[16]

    return bf_art, bf_vein


def whole_body(bf_art, bf_vein, bf_ava_hand, bf_ava_foot):
    """Create matrix to calculate heat exchange by blood flow between segments.

    [W/K]
    """

    def flow(up, down, bloodflow):
        arr = np.zeros((NUM_NODES, NUM_NODES))
        arr[down, up] = (
            1.067 * bloodflow
        )  # Coefficient = 1.067 [Wh/L.K] Change units [L/h] to [W/K]
        return arr

    arr83 = np.zeros((NUM_NODES, NUM_NODES))
    # Matrix offsets of segments
    CB = IDICT["CB"]
    head = IDICT["head"]["artery"]
    neck = IDICT["neck"]["artery"]
    chest = IDICT["chest"]["artery"]
    back = IDICT["back"]["artery"]
    pelvis = IDICT["pelvis"]["artery"]
    left_shoulder = IDICT["left_shoulder"]["artery"]
    left_arm = IDICT["left_arm"]["artery"]
    left_hand = IDICT["left_hand"]["artery"]
    right_shoulder = IDICT["right_shoulder"]["artery"]
    right_arm = IDICT["right_arm"]["artery"]
    right_hand = IDICT["right_hand"]["artery"]
    left_thigh = IDICT["left_thigh"]["artery"]
    left_leg = IDICT["left_leg"]["artery"]
    left_foot = IDICT["left_foot"]["artery"]
    right_thigh = IDICT["right_thigh"]["artery"]
    right_leg = IDICT["right_leg"]["artery"]
    right_foot = IDICT["right_foot"]["artery"]

    arr83 += flow(CB, neck, bf_art[1])  # CB to neck.art
    arr83 += flow(neck, head, bf_art[0])  # neck.art to head.art
    arr83 += flow(head + 1, neck + 1, bf_vein[0])  # head.vein to neck.vein
    arr83 += flow(neck + 1, CB, bf_vein[1])  # neck.vein to CB

    arr83 += flow(CB, chest, bf_art[2])  # CB to chest.art
    arr83 += flow(chest + 1, CB, bf_vein[2])  # chest.vein to CB

    arr83 += flow(CB, back, bf_art[3])  # CB to back.art
    arr83 += flow(back + 1, CB, bf_vein[3])  # back.vein to CB

    arr83 += flow(CB, pelvis, bf_art[4])  # CB to pelvis.art
    arr83 += flow(pelvis + 1, CB, bf_vein[4])  # pelvis.vein to CB

    arr83 += flow(CB, left_shoulder, bf_art[5])  # CB to left_shoulder.art
    arr83 += flow(
        left_shoulder,
        left_arm,
        bf_art[6],
    )  # left_shoulder.art to left_arm.art
    arr83 += flow(left_arm, left_hand, bf_art[7])  # left_arm.art to left_hand.art
    arr83 += flow(
        left_hand + 1,
        left_arm + 1,
        bf_vein[7],
    )  # left_hand.vein to left_arm.vein
    arr83 += flow(
        left_arm + 1,
        left_shoulder + 1,
        bf_vein[6],
    )  # left_arm.vein to left_shoulder.vein
    arr83 += flow(left_shoulder + 1, CB, bf_vein[5])  # left_shoulder.vein to CB
    arr83 += flow(
        left_hand + 2,
        left_arm + 2,
        bf_ava_hand,
    )  # left_hand.sfvein to left_arm.sfvein
    arr83 += flow(
        left_arm + 2,
        left_shoulder + 2,
        bf_ava_hand,
    )  # left_arm.sfvein to left_shoulder.sfvein
    arr83 += flow(left_shoulder + 2, CB, bf_ava_hand)  # left_shoulder.sfvein to CB

    arr83 += flow(CB, right_shoulder, bf_art[8])  # CB to right_shoulder.art
    arr83 += flow(
        right_shoulder,
        right_arm,
        bf_art[9],
    )  # right_shoulder.art to right_arm.art
    arr83 += flow(right_arm, right_hand, bf_art[10])  # right_arm.art to right_hand.art
    arr83 += flow(
        right_hand + 1,
        right_arm + 1,
        bf_vein[10],
    )  # right_hand.vein to right_arm.vein
    arr83 += flow(
        right_arm + 1,
        right_shoulder + 1,
        bf_vein[9],
    )  # right_arm.vein to right_shoulder.vein
    arr83 += flow(right_shoulder + 1, CB, bf_vein[8])  # right_shoulder.vein to CB
    arr83 += flow(
        right_hand + 2,
        right_arm + 2,
        bf_ava_hand,
    )  # right_hand.sfvein to right_arm.sfvein
    arr83 += flow(
        right_arm + 2,
        right_shoulder + 2,
        bf_ava_hand,
    )  # right_arm.sfvein to right_shoulder.sfvein
    arr83 += flow(right_shoulder + 2, CB, bf_ava_hand)  # right_shoulder.sfvein to CB

    arr83 += flow(pelvis, left_thigh, bf_art[11])  # pelvis to left_thigh.art
    arr83 += flow(left_thigh, left_leg, bf_art[12])  # left_thigh.art to left_leg.art
    arr83 += flow(left_leg, left_foot, bf_art[13])  # left_leg.art to left_foot.art
    arr83 += flow(
        left_foot + 1,
        left_leg + 1,
        bf_vein[13],
    )  # left_foot.vein to left_leg.vein
    arr83 += flow(
        left_leg + 1,
        left_thigh + 1,
        bf_vein[12],
    )  # left_leg.vein to left_thigh.vein
    arr83 += flow(left_thigh + 1, pelvis + 1, bf_vein[11])  # left_thigh.vein to pelvis
    arr83 += flow(
        left_foot + 2,
        left_leg + 2,
        bf_ava_foot,
    )  # left_foot.sfvein to left_leg.sfvein
    arr83 += flow(
        left_leg + 2,
        left_thigh + 2,
        bf_ava_foot,
    )  # left_leg.sfvein to left_thigh.sfvein
    arr83 += flow(left_thigh + 2, pelvis + 1, bf_ava_foot)  # left_thigh.vein to pelvis

    arr83 += flow(pelvis, right_thigh, bf_art[14])  # pelvis to right_thigh.art
    arr83 += flow(
        right_thigh,
        right_leg,
        bf_art[15],
    )  # right_thigh.art to right_leg.art
    arr83 += flow(right_leg, right_foot, bf_art[16])  # right_leg.art to right_foot.art
    arr83 += flow(
        right_foot + 1,
        right_leg + 1,
        bf_vein[16],
    )  # right_foot.vein to right_leg.vein
    arr83 += flow(
        right_leg + 1,
        right_thigh + 1,
        bf_vein[15],
    )  # right_leg.vein to right_thigh.vein
    arr83 += flow(
        right_thigh + 1,
        pelvis + 1,
        bf_vein[14],
    )  # right_thigh.vein to pelvis
    arr83 += flow(
        right_foot + 2,
        right_leg + 2,
        bf_ava_foot,
    )  # right_foot.sfvein to right_leg.sfvein
    arr83 += flow(
        right_leg + 2,
        right_thigh + 2,
        bf_ava_foot,
    )  # right_leg.sfvein to right_thigh.sfvein
    arr83 += flow(
        right_thigh + 2,
        pelvis + 1,
        bf_ava_foot,
    )  # right_thigh.vein to pelvis

    return arr83


def remove_body_name(text):
    """Remove the body name from the parameter name.

    Parameters
    ----------
    text : str
        Parameter name

    Returns
    -------
    rtext : str
        Parameter name removed the body name.
    removed : str
        The removed body name

    """
    rtext = text
    removed = None

    # Remove the body part name from the parameter name
    for bn in JOS3BodyParts.get_attribute_names():
        if bn in text:
            rtext = rtext.replace(
                bn,
                "",
            )  # Remove the body part name from the parameter name
            rtext = rtext.removesuffix("_")  # Remove the trailing underscore
            removed = bn  # Store the removed body part name
            break
    return rtext, removed



from __future__ import annotations

import math

import numpy as np

from pythermalcomfort.classes_input import HIInputs
from pythermalcomfort.classes_return import HI


def heat_index_lu(
    tdb: float | list[float],
    rh: float | list[float],
    round_output: bool = True,
) -> HI:
    """Calculate the Heat Index (HI) in accordance with the Lu and Romps (2022) model [lu]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    rh : float or list of floats
        Relative humidity, [%].
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    HI
        A dataclass containing the Heat Index. See :py:class:`~pythermalcomfort.classes_return.HI` for more details.
        To access the `hi` value, use the `hi` attribute of the returned `HI` instance, e.g., `result.hi`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import heat_index_lu

        result = heat_index_lu(tdb=25, rh=50)
        print(result.hi)  # 25.9

    """
    # Validate inputs using the HeatIndexInputs class
    HIInputs(
        tdb=tdb,
        rh=rh,
        round_output=round_output,
        limit_inputs=False,
    )

    tdb = np.array(tdb)
    rh = np.array(rh)

    hi = _lu_heat_index_vectorized(tdb + 273.15, rh / 100) - 273.15

    if round_output:
        hi = np.around(hi, 1)

    return HI(hi=hi)


# combining the two functions find_eqvar and find_T
@np.vectorize
def _lu_heat_index_vectorized(tdb, rh):  # Thermodynamic parameters
    t_c_k = 273.16  # K
    p_triple_point = 611.65  # Pa
    e0v = 2.3740e6  # J/kg
    e0s = 0.3337e6  # J/kg
    rgasa = 287.04  # J/kg/K
    rgasv = 461.0  # J/kg/K
    cva = 719.0  # J/kg/K
    cvv = 1418.0  # J/kg/K
    cvl = 4119.0  # J/kg/K
    cvs = 1861.0  # J/kg/K
    cpa = cva + rgasa
    cpv = cvv + rgasv

    # The saturation vapor pressure
    def pv_star(t):
        if t == 0.0:
            return 0.0
        if t < t_c_k:
            return (
                p_triple_point
                * (t / t_c_k) ** ((cpv - cvs) / rgasv)
                * math.exp(
                    (e0v + e0s - (cvv - cvs) * t_c_k) / rgasv * (1.0 / t_c_k - 1.0 / t),
                )
            )
        return (
            p_triple_point
            * (t / t_c_k) ** ((cpv - cvl) / rgasv)
            * math.exp(
                (e0v - (cvv - cvl) * t_c_k) / rgasv * (1.0 / t_c_k - 1.0 / t),
            )
        )

    # The latent heat of vaporization of water
    def latent_heat_vap(t):
        return e0v + (cvv - cvl) * (t - t_c_k) + rgasv * t

    # Thermo-regulatory parameters
    sigma = 5.67e-8  # W/m^2/K^4 , Stefan-Boltzmann constant
    epsilon = 0.97  # emissivity of surface, steadman1979
    mass = 83.6  # kg, mass of average US adults, fryar2018
    height = 1.69  # m, height of average US adults, fryar2018
    area = 0.202 * (mass**0.425) * (height**0.725)  # m^2, DuBois formula, parson2014
    cpc = 3492.0  # J/kg/K, specific heat capacity of core, gagge1972
    hc_core = mass * cpc / area  # heat capacity of core
    r = 124.0  # Pa/K, Zf/rf, steadman1979
    q = 180.0  # W/m^2, metabolic rate per skin area, steadman1979
    phi_salt = 0.9  # vapor saturation pressure level of saline solution, steadman1979
    t_cr = 310.0  # K, core temperature, steadman1979
    p_cr = phi_salt * pv_star(t_cr)  # core vapor pressure
    lat_heat = latent_heat_vap(310.0)  # latent heat of vaporization at 310 K
    p = 1.013e5  # Pa, atmospheric pressure
    eta = 1.43e-6  # kg/J, "inhaled mass" / "metabolic rate", steadman1979
    pa0 = 1.6e3  # Pa, reference air vapor pressure in regions III, IV, V, VI, steadman1979

    za = (
        60.6 / 17.4
    )  # Pa m^2/W, mass transfer resistance through air, exposed part of skin
    za_bar = (
        60.6 / 11.6
    )  # Pa m^2/W, mass transfer resistance through air, clothed part of skin
    za_un = (
        60.6 / 12.3
    )  # Pa m^2/W, mass transfer resistance through air, when being naked

    # tolerance and maximum iteration for the root solver
    tol = 1e-8
    tol_t = 1e-8
    max_iter = 100

    # Thermo-regulatory functions
    def qv(ta, pa):  # respiratory heat loss, W/m^2
        return (
            eta * q * (cpa * (t_cr - ta) + lat_heat * rgasa / (p * rgasv) * (p_cr - pa))
        )

    def zs(rs):  # mass transfer resistance through skin, Pa m^2/W
        return 52.1 if rs == 0.0387 else 6.0e8 * rs**5

    def ra(
        ts,
        ta,
    ):  # heat transfer resistance through air, exposed part of skin, K m^2/W
        hc = 17.4
        phi_rad = 0.85
        hr = epsilon * phi_rad * sigma * (ts**2 + ta**2) * (ts + ta)
        return 1.0 / (hc + hr)

    def ra_bar(
        tf,
        ta,
    ):  # heat transfer resistance through air, clothed part of skin, K m^2/W
        hc = 11.6
        phi_rad = 0.79
        hr = epsilon * phi_rad * sigma * (tf**2 + ta**2) * (tf + ta)
        return 1.0 / (hc + hr)

    def ra_un(
        ts,
        ta,
    ):  # heat transfer resistance through air, when being naked, K m^2/W
        hc = 12.3
        phi_rad = 0.80
        hr = epsilon * phi_rad * sigma * (ts**2 + ta**2) * (ts + ta)
        return 1.0 / (hc + hr)

    # Given air temperature and relative humidity, returns the equivalent variables
    def find_eq_var(ta, _rh):
        pa = _rh * pv_star(ta)  # air vapor pressure
        rs = 0.0387  # m^2K/W, heat transfer resistance through skin
        phi = 0.84  # covering fraction
        d_tc_dt = 0.0  # K/s, rate of change in Tc
        m = (p_cr - pa) / (zs(rs) + za)
        m_bar = (p_cr - pa) / (zs(rs) + za_bar)
        ts = solve(
            lambda ts: (ts - ta) / ra(ts, ta)
            + (p_cr - pa) / (zs(rs) + za)
            - (t_cr - ts) / rs,
            max(0.0, min(t_cr, ta) - rs * abs(m)),
            max(t_cr, ta) + rs * abs(m),
            tol,
            max_iter,
        )
        tf = solve(
            lambda tf: (tf - ta) / ra_bar(tf, ta)
            + (p_cr - pa) / (zs(rs) + za_bar)
            - (t_cr - tf) / rs,
            max(0.0, min(t_cr, ta) - rs * abs(m_bar)),
            max(t_cr, ta) + rs * abs(m_bar),
            tol,
            max_iter,
        )
        flux1 = (
            q - qv(ta, pa) - (1.0 - phi) * (t_cr - ts) / rs
        )  # C*dTc/dt when rf=Zf=\inf
        flux2 = (
            q - qv(ta, pa) - (1.0 - phi) * (t_cr - ts) / rs - phi * (t_cr - tf) / rs
        )  # C*dTc/dt when rf=Zf=0
        if flux1 <= 0.0:  # region I
            eq_var_name = "phi"
            phi = 1.0 - (q - qv(ta, pa)) * rs / (t_cr - ts)
            rf = float("inf")
        elif flux2 <= 0.0:  # region II&III
            eq_var_name = "rf"
            ts_bar = (
                t_cr - (q - qv(ta, pa)) * rs / phi + (1.0 / phi - 1.0) * (t_cr - ts)
            )
            tf = solve(
                lambda tf: (tf - ta) / ra_bar(tf, ta)
                + (p_cr - pa)
                * (tf - ta)
                / ((zs(rs) + za_bar) * (tf - ta) + r * ra_bar(tf, ta) * (ts_bar - tf))
                - (t_cr - ts_bar) / rs,
                ta,
                ts_bar,
                tol,
                max_iter,
            )
            rf = ra_bar(tf, ta) * (ts_bar - tf) / (tf - ta)
        else:  # region IV,V,VI
            rf = 0.0
            flux3 = (
                q
                - qv(ta, pa)
                - (t_cr - ta) / ra_un(t_cr, ta)
                - (phi_salt * pv_star(t_cr) - pa) / za_un
            )
            if flux3 < 0.0:  # region IV,V
                ts = solve(
                    lambda ts: (ts - ta) / ra_un(ts, ta)
                    + (p_cr - pa) / (zs((t_cr - ts) / (q - qv(ta, pa))) + za_un)
                    - (q - qv(ta, pa)),
                    0.0,
                    t_cr,
                    tol,
                    max_iter,
                )
                rs = (t_cr - ts) / (q - qv(ta, pa))
                eq_var_name = "rs"
                ps = p_cr - (p_cr - pa) * zs(rs) / (zs(rs) + za_un)
                if ps > phi_salt * pv_star(ts):  # region V
                    ts = solve(
                        lambda ts: (ts - ta) / ra_un(ts, ta)
                        + (phi_salt * pv_star(ts) - pa) / za_un
                        - (q - qv(ta, pa)),
                        0.0,
                        t_cr,
                        tol,
                        max_iter,
                    )
                    rs = (t_cr - ts) / (q - qv(ta, pa))
                    eq_var_name = "rs*"
            else:  # region VI
                rs = 0.0
                eq_var_name = "d_tc_dt"
                d_tc_dt = (1.0 / hc_core) * flux3
        return [eq_var_name, phi, rf, rs, d_tc_dt]

    # given the equivalent variable, find the Heat Index
    def find_t(eq_var_name, eq_var):
        if eq_var_name == "phi":
            t = solve(
                lambda t: find_eq_var(t, 1.0)[1] - eq_var,
                0.0,
                240.0,
                tol_t,
                max_iter,
            )
            _region = "I"
        elif eq_var_name == "rf":
            t = solve(
                lambda t: find_eq_var(t, min(1.0, pa0 / pv_star(t)))[2] - eq_var,
                230.0,
                300.0,
                tol_t,
                max_iter,
            )
            _region = "II" if pa0 > pv_star(t) else "III"
        elif eq_var_name == "rs" or eq_var_name == "rs*":
            t = solve(
                lambda T: find_eq_var(T, pa0 / pv_star(T))[3] - eq_var,
                295.0,
                350.0,
                tol_t,
                max_iter,
            )
            _region = "IV" if eq_var_name == "rs" else "V"
        else:
            t = solve(
                lambda T: find_eq_var(T, pa0 / pv_star(T))[4] - eq_var,
                340.0,
                1000.0,
                tol_t,
                max_iter,
            )
            _region = "VI"
        return t, _region

    def solve(f, x1, x2, _tol, _max_iter):
        a = x1
        b = x2
        fa = f(a)
        fb = f(b)
        if fa * fb > 0.0:
            raise SystemExit("wrong initial interval in the root solver")
        for i in range(_max_iter):
            c = (a + b) / 2.0
            fc = f(c)
            if fb * fc > 0.0:
                b = c
                fb = fc
            else:
                a = c
            if abs(a - b) < _tol:
                return c
            if i == _max_iter - 1:
                raise SystemExit("reaching maximum iteration in the root solver")

    dic = {"phi": 1, "rf": 2, "rs": 3, "rs*": 3, "d_tc_dt": 4}
    eq_vars = find_eq_var(tdb, rh)
    hi, region = find_t(eq_vars[0], eq_vars[dic[eq_vars[0]]])
    if tdb == 0.0:
        hi = 0.0
    return hi



from __future__ import annotations

import json
from collections.abc import Callable
from enum import Enum
from typing import Any

import numpy as np
import pytest
import requests

# The conftest.py file serves as a means of providing fixtures for an entire directory.
# Fixtures defined in a conftest.py can be used by any test in that package
# without needing to import them (pytest will automatically discover them).


unit_test_data_prefix = "https://raw.githubusercontent.com/FedericoTartarini/validation-data-comfort-models/main/"


class Urls(Enum):
    """Enum for URLs of test data files."""

    ADAPTIVE_EN = "ts_adaptive_en.json"
    ADAPTIVE_ASHRAE = "ts_adaptive_ashrae.json"
    A_PMV = "ts_a_pmv.json"
    TWO_NODES = "ts_two_nodes_gagge.json"
    SOLAR_GAIN = "ts_solar_gain.json"
    ANKLE_DRAFT = "ts_ankle_draft.json"
    PHS = "ts_phs.json"
    E_PMV = "ts_e_pmv.json"
    AT = "ts_at.json"
    ATHB = "ts_athb.json"
    CLO_TOUT = "ts_clo_tout.json"
    COOLING_EFFECT = "ts_cooling_effect.json"
    VERTICAL_TMP_GRAD_PPD = "ts_vertical_tmp_grad_ppd.json"
    WBGT = "ts_wbgt.json"
    HEAT_INDEX = "ts_heat_index.json"
    NET = "ts_net.json"
    PMV_PPD = "ts_pmv_ppd.json"
    PMV = "ts_pmv.json"
    SET = "ts_set.json"
    HUMIDEX = "ts_humidex.json"
    USE_FANS_HEATWAVES = "ts_use_fans_heatwaves.json"
    UTCI = "ts_utci.json"
    WIND_CHILL = "ts_wind_chill.json"
    PET_STEADY = "ts_pet_steady.json"
    DISCOMFORT_INDEX = "ts_discomfort_index.json"


@pytest.fixture
def get_test_url() -> Callable[[str], str]:
    """Return a function that builds the full test data URL for a given model name.

    Returns
    -------
    Callable[[str], str]
        Callable accepting a model_name string and returning the complete URL.

    """

    def _get_test_url(model_name: str) -> str:
        """Construct the test URL for the given model name or return empty on failure."""
        try:
            return unit_test_data_prefix + Urls[model_name.upper()].value
        except KeyError:
            return ""

    return _get_test_url


@pytest.fixture
def retrieve_data() -> Callable[[str], dict[str, Any] | None]:
    """Return a function that fetches JSON data from a URL.

    Returns
    -------
    Callable[[str], Optional[Dict[str, Any]]]
        A callable accepting a URL string and returning a dict or None.

    """

    def _retrieve_data(url: str) -> dict[str, Any] | None:
        try:
            response = requests.get(url, timeout=10)
            if response.status_code == 200:
                return json.loads(response.text)
            response.raise_for_status()
        except requests.RequestException as e:
            message = f"Error fetching data from {url}: {e}"
            raise RuntimeError(message) from e
        return None

    return _retrieve_data


def is_equal(a, b, tolerance=1e-6) -> bool:
    """Compare two values for equality with a specified tolerance."""
    if isinstance(a, np.ndarray):
        if not isinstance(b, np.ndarray):
            b = np.array(b, dtype=a.dtype)
        if a.dtype.kind in "UOS":  # U = unicode, O = objects, S = string
            return np.array_equal(a, b)
        b = np.where(b is None, np.nan, b)  # Replace None with np.nan
        # Return True if arrays are close enough, including handling of NaN values
        return np.allclose(a, b, atol=tolerance, equal_nan=True)
    if (a is None and np.isnan(b)) or (b is None and np.isnan(a)):
        return True
    if isinstance(a, int | float) and isinstance(b, int | float):
        # Compare scalar values with tolerance
        return np.isclose(a, b, atol=tolerance)
    return a == b


def retrieve_reference_table(get_test_url: str, retrieve_data, url_name):
    """Retrieve the reference table for a given model from the test data URL."""
    reference_table = retrieve_data(get_test_url(url_name))
    if reference_table is None:
        pytest.fail(f"Failed to retrieve reference table for {url_name.lower()}")
    return reference_table


def validate_result(result: dict, expected_output: dict, tolerance: dict) -> None:
    """Validate the result of a function against expected output with a tolerance.

    Parameters
    ----------
    result : dict
        this is the result of the function that is being tested
    expected_output :dict
        this is the expected output of the function that is being tested
    tolerance : dict
        this is the tolerance that is used to compare the result with the
        expected output

    Returns
    -------
    None

    """
    for key in expected_output:
        _expected_output = expected_output[key]

        # some functions return a dictionary or class with multiple values while
        # others return a single value
        # TODO remove this once all the functions return a dictionary or class
        try:
            _result = result[key]
        except (IndexError, KeyError, TypeError):
            _result = result

        # if the key is not in the tolerance dictionary we set the tolerance to 1e-6
        _tolerance = tolerance.get(key, 1e-06)

        try:
            assert is_equal(_result, _expected_output, _tolerance)
        except Exception as e:
            message = f"Expected {_expected_output}, got {_result}\nError: {e!s}"
            raise RuntimeError(message) from e



from __future__ import annotations

import math

import numpy as np

from pythermalcomfort.classes_input import GaggeTwoNodesJiInputs
from pythermalcomfort.classes_return import GaggeTwoNodesJi
from pythermalcomfort.utilities import Postures


def two_nodes_gagge_ji(
    tdb: float | list[float],
    tr: float | list[float],
    v: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    vapor_pressure: float | list[float],
    wme: float | list[float] = 0,
    body_surface_area: float | list[float] = 1.8258,
    p_atm: float | list[float] = 101325,
    position: str = Postures.sitting.value,
    **kwargs,
) -> GaggeTwoNodesJi:
    """Two-node model for older people under hot and cold exposures proposed by Ji et al. [Ji2022]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    tr : float or list of floats
        Mean radiant temperature, [°C].
    v : float or list of floats
        Air speed, [m/s].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].
    vapor_pressure : float or list of floats
        Vapor pressure, [torr].

        .. note::
            Vapor pressure can be calculated using the relative humidity and the saturation
            vapor pressure, which can be computed using the function
            :py:meth:`pythermalcomfort.utilities.p_sat_torr`.

    wme : float or list of floats
        External work, [met]. Defaults to 0.
    body_surface_area : float or list of floats
        Body surface area, [m2]. Defaults to 1.8258.

        .. note::
            Body surface area can be calculated with weight and height using the function
            :py:meth:`pythermalcomfort.utilities.body_surface_area`.

    p_atm : float or list of floats
        Atmospheric pressure, [Pa]. Defaults to 101325.

    position : str, optional
        Select either "sitting" or "standing". Defaults to "sitting".

    Other Parameters
    ----------------
    body_weight : float, optional
        Body weight, [kg]. Defaults to 70 kg.
    length_time_simulation : int, optional
        Length of the simulation, [minutes]. Defaults to 120 minutes.
    initial_skin_temp : float, optional
        Initial skin temperature, [°C]. Defaults to 36.8 °C.
    initial_core_temp : float, optional
        Initial core temperature, [°C]. Defaults to 36.49 °C.
    acclimatized : bool, optional
        If True, the model assumes the person is acclimatized to heat. Defaults to True.

    Returns
    -------
    GaggeTwoNodesJi
        A dataclass containing the simulated core and skin temperatures over time. See :py:class:`~pythermalcomfort.classes_return.GaggeTwoNodesJi` for more details.
        To access the individual attributes, use the corresonding attrubute of the returned
        instance, e.g. `result.t_core` or `result.t_skin`.

    Example
    -------
    .. code-block:: python

        from pythermalcomfort.models import two_nodes_gagge_ji
        from pythermalcomfort.utilities import body_surface_area
        from pythermalcomfort.utilities import p_sat_torr

        rh = 20
        vapor_pressure = rh * p_sat_torr(tdb=36.5) / 100

        result = two_nodes_gagge_ji(
            tdb=36.5,
            tr=36.5,
            v=0.25,
            met=0.95,
            clo=0.1,
            vapor_pressure=vapor_pressure,
            wme=0,
            body_surface_area=body_surface_area(weight=80.1, height=1.8),
            p_atm=101325,
            position="sitting",
            acclimatized=True,
            body_weight=80.1,
            length_time_simulation=120,
        )

    """
    body_weight = kwargs.pop("body_weight", 70)
    length_time_simulation = kwargs.pop("length_time_simulation", 120)
    initial_skin_temp = kwargs.pop("initial_skin_temp", 36.8)
    initial_core_temp = kwargs.pop("initial_core_temp", 36.49)
    acclimatized = kwargs.pop("acclimatized", True)

    if kwargs:
        error_msg = f"Unexpected keyword arguments: {list(kwargs.keys())}"
        raise TypeError(error_msg)

    tdb = np.array(tdb)
    tr = np.array(tr)
    v = np.array(v)
    met = np.array(met)
    clo = np.array(clo)
    vapor_pressure = np.array(vapor_pressure)
    wme = np.array(wme)
    body_surface_area = np.array(body_surface_area)
    p_atm = np.array(p_atm)

    # Validate inputs
    GaggeTwoNodesJiInputs(
        tdb=tdb,
        tr=tr,
        v=v,
        met=met,
        clo=clo,
        vapor_pressure=vapor_pressure,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atm=p_atm,
        position=position,
    )

    excluded_params = {
        "position",
        "acclimatized",
        "body_weight",
        "length_time_simulation",
        "initial_skin_temp",
        "initial_core_temp",
    }

    results_array_of_dicts = np.vectorize(
        _two_nodes_ji_optimized,
        excluded=excluded_params,
        otypes=[object],
    )(
        tdb=tdb,
        tr=tr,
        v=v,
        met=met,
        clo=clo,
        vapor_pressure=vapor_pressure,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atm=p_atm,
        position=position,
        acclimatized=acclimatized,
        body_weight=body_weight,
        length_time_simulation=length_time_simulation,
        initial_skin_temp=initial_skin_temp,
        initial_core_temp=initial_core_temp,
    )

    output_data = {}
    if results_array_of_dicts.ndim == 0:
        # scalar inputs
        output_data = results_array_of_dicts.item()
    elif results_array_of_dicts.size == 0:
        # empty input arrays
        output_data = {"t_core": [], "t_skin": []}
    else:
        # multiple simulations were run with array inputs
        # t_core and t_skin are arrays of arrays, each array corresponds to a simulation
        first_result_dict = results_array_of_dicts[0]
        for key in first_result_dict.keys():
            output_data[key] = []

        for res_dict in results_array_of_dicts:
            for key, value in res_dict.items():
                # creates a list of NumPy arrays for each key
                output_data[key].append(value)

    return GaggeTwoNodesJi(**output_data)


def _two_nodes_ji_optimized(
    tdb,
    tr,
    v,
    met,
    clo,
    vapor_pressure,
    wme,
    body_surface_area,
    p_atm,
    position,
    acclimatized,
    body_weight,
    length_time_simulation,
    initial_skin_temp,
    initial_core_temp,
):
    # Initial variables as defined in the ASHRAE 55-2020
    air_speed = max(v, 0.1)
    met_factor = 58.2  # met conversion factor
    sbc = 0.000000056697  # Stefan-Boltzmann constant (W/m2K4)

    # These values are taken from Ji, 2022
    c_sw = 170  # driving coefficient for regulatory sweating
    c_dil = 50  # driving coefficient for vasodilation
    c_str = 0.75  # driving coefficient for vasoconstriction
    a_cof = 0.2  # coefficient in sweat rate

    # Max and min values defined by Ji, 2022
    min_skin_blood_flow = 0.75  # min SBF for older people
    max_skin_blood_flow = 63  # max SBF for older people
    max_sweating_rate_factor = 0.9  # 90% sweating efficiency
    evap_sweating_reg_max = 400  # W/m2

    # Other coefficients defined by Ji, 2022
    c_de = 0.6  # attenuation coefficient of vasodilation
    c_ce = 0.5  # attenuation coefficient of vasoconstriction
    c_swe = 1  # sweat attenuation coefficient
    c_she = 1  # shivering attenuation coefficient
    cof_scs = 19.4  # added shivering coefficient
    cof_sc = 50  # added shivering coefficient
    cof_ss = 0.5  # added shivering coefficient

    # Neutral/trigger values for different systems (in older people)
    # Taken from Ji, 2022
    t_cr0_dil = 37.3  # vasodilation threshold
    t_sk0_cons = 33.25  # vasoconstriction threshold
    t_cr0_sw = 37  # core sweating threshold
    t_sk0_sw = 34.3  # skin sweating threshold
    t_cr0_sh = 36.7  # shivering threshold

    alfa = 0.1
    skin_blood_flow_neutral = 6.3  # as defined by Ji, 2021

    t_skin = initial_skin_temp
    t_core = initial_core_temp
    m_bl = skin_blood_flow_neutral

    # initialize some variables
    e_skin = 0.1 * met  # total evaporative heat loss, W
    w = 0.06  # skin wettedness

    pressure_in_atmospheres = p_atm / 101325
    n_simulation = 0

    r_clo = 0.155 * clo  # thermal resistance of clothing, K*m2/W
    # increase in body surface area due to clothing
    f_a_cl = 1.0 + 0.2 * clo if clo < 0.5 else 1.05 + 0.1 * clo
    lr = 2.2 / pressure_in_atmospheres  # Lewis relation
    m = met * met_factor  # metabolic rate in W/m2

    i_cl = 1.0  # permeation efficiency of water vapor naked skin
    if clo > 0:
        i_cl = 0.45  # permeation efficiency of water vapor through the clothing layer

    # Set w_max and the max sweating rate, taken from emails with Dr Laouadi
    w_max = 0.85
    if acclimatized:
        evap_sweating_reg_max = 1.25 * evap_sweating_reg_max
        w_max = 1
    m_rsw_max = evap_sweating_reg_max / 0.68
    m_rsw_max = m_rsw_max * max_sweating_rate_factor

    if not w_max:  # if the user did not pass a value of w_max
        # W_max should be fixed to either 0.85 or 1 (for acclimatized people);  See ISO PHS model and other related standard
        w_max = 1

    # Heat transfer coefficients
    h_cc = 3  # initial value - convective heat transfer coefficient
    h_r = 4.7  # initial value - linearized radiative heat transfer coefficient

    skin_temp_hist = []
    core_temp_hist = []

    while n_simulation < length_time_simulation:
        n_simulation += 1

        iteration_limit = 150  # for following while loop

        h_t = (
            h_r + h_cc
        )  # sum of convective and radiant heat transfer coefficient W/(m2*K)
        r_a = 1.0 / (f_a_cl * h_t)  # resistance of air layer to dry heat
        t_op = (h_r * tr + h_cc * tdb) / h_t  # operative temperature

        # t_cl temperature of the outer surface of clothing
        t_cl = (r_a * t_skin + r_clo * t_op) / (r_a + r_clo)  # initial guess
        n_iterations = 0
        tc_converged = False

        while not tc_converged:
            if position == Postures.sitting.value:
                # 0.7 ratio between radiation area of the body and the body area
                h_r = 4.0 * 0.97 * sbc * ((t_cl + tr) / 2.0 + 273.15) ** 3.0 * 0.7
            else:  # if standing
                # 0.77 ratio between radiation area of the body and the body area
                h_r = 4.0 * 0.97 * sbc * ((t_cl + tr) / 2.0 + 273.15) ** 3.0 * 0.77
            h_t = h_r + h_cc
            r_a = 1.0 / (f_a_cl * h_t)
            t_op = (h_r * tr + h_cc * tdb) / h_t
            t_cl_new = (r_a * t_skin + r_clo * t_op) / (r_a + r_clo)
            if abs(t_cl_new - t_cl) < 0.01:
                tc_converged = True
            t_cl = t_cl_new
            n_iterations += 1

            if n_iterations > iteration_limit:
                raise StopIteration("Max iterations exceeded")

        # Convective heat transfer coefficient based on clothing
        d_tcl_air = t_cl - tdb  # difference between clothing and air temperature
        if air_speed < 0.2:
            # Free convection. Gao et al 2019: Formulation of human body heat transfer coefficient...
            if d_tcl_air > 0:
                # Upward flow
                h_cc = 2.5 * d_tcl_air**0.16
            else:
                # Downward flow
                h_cc = 2.5 * math.fabs(d_tcl_air) ** 0.41
        else:
            # Forced convection
            # Original formula from SET code
            h_cc = 8.6 * air_speed**0.53

        q_sensible = (t_skin - t_op) / (r_a + r_clo)  # total sensible heat loss, W

        # respiration
        q_res = (
            0.0023 * m * (44.0 - vapor_pressure)
        )  # latent heat loss due to respiration
        c_res = 0.0014 * m * (34.0 - tdb)  # sensible convective heat loss respiration

        # hf_cs rate of energy transport between core and skin, W
        # 5.28 is the average body tissue conductance in W/(m2 C)
        # 1.163 is the thermal capacity of blood in Wh/(L C)
        hf_cs = (t_core - t_skin) * (5.28 + 1.163 * m_bl)
        s_core = m - hf_cs - q_res - c_res - wme  # rate of energy storage in the core
        s_skin = hf_cs - q_sensible - e_skin  # rate of energy storage in the skin
        tc_sk = 0.97 * alfa * body_weight  # thermal capacity skin
        tc_cr = 0.97 * (1 - alfa) * body_weight  # thermal capacity core
        d_t_sk = (s_skin * body_surface_area) / (
            tc_sk * 60.0
        )  # rate of change skin temperature °C per minute
        d_t_cr = (s_core * body_surface_area) / (
            tc_cr * 60.0
        )  # rate of change core temperature °C per minute
        t_skin = t_skin + d_t_sk
        t_core = t_core + d_t_cr

        #
        # Main equations used in Ji, 2022
        #

        # Calculate skin blood flow (SBF)
        t_cr_dil = max(0, t_core - t_cr0_dil)  # dilation trigger
        t_sk_cons = max(0, t_sk0_cons - t_skin)  # constriction trigger
        m_bl = (skin_blood_flow_neutral + c_de * c_dil * t_cr_dil) / (
            1 + c_ce * c_str * t_sk_cons
        )  # sbf
        # Set min/max skin blood flow values
        m_bl = min(max_skin_blood_flow, m_bl)
        m_bl = max(min_skin_blood_flow, m_bl)

        # Calculate sweat triggers
        t_sk_sw = max(0, t_skin - t_sk0_sw)
        t_cr_sw = max(0, t_core - t_cr0_sw)

        # Update alfa (for next cycle)
        alfa = 0.0417737 + 0.7451832 / (m_bl + 0.5854417)

        # Sweat rate (SWR) - accounting for the metabolic rate
        m_rsw = (
            c_swe
            * c_sw
            * ((1 - alfa) * t_cr_sw + (alfa + a_cof) * t_sk_sw)
            * math.exp(t_sk_sw / 10.7)
        )
        # Apply max sweating rate
        m_rsw = min(m_rsw, m_rsw_max)
        # Energy lost via evaporation
        e_rsw = 0.68 * m_rsw  # heat lost by vaporization sweat

        r_e_cl = r_clo / (lr * i_cl)  # evaporative resistance of clothing in K*m2*atm/W
        r_e_a = 1 / (
            lr * f_a_cl * h_cc
        )  # evaporative resistance of air layer in K*m2*atm/W
        r_total = r_e_cl + r_e_a  # total evaporative resistance in K*m2*atm/W

        e_max = (
            math.exp(18.6686 - 4030.183 / (t_skin + 235.0)) - vapor_pressure
        ) / r_total
        p_rsw = e_rsw / e_max  # ratio heat loss sweating to max heat loss sweating

        # calculate first the sweat efficiency using the skin WETNESS (assuming all sweat is converted to evaporative heat).
        # Note skin wetness is different than skin WETTEDNESS which assumes the correct evaporation efficiency
        he_n = (
            1 / r_total
        )  # coefficient of the evaporative heat transfer (inverse of the total evaporative heat resistance; W/(K*m2*Pa))
        wettedness_dif = 1 / (2 + 2.46 * he_n)
        wp = (
            wettedness_dif + (1 - wettedness_dif) * p_rsw
        )  # skin wetness (between 0.06 and 1)

        # Calculate skin wettedness, by taking the ISO PHS evaporation efficiency formula: eff = 1 - 0.5 * w**2
        w = (math.sqrt(2 * wp**2 + 1) - 1) / wp

        # Limit w
        # print("Skin wetness exceeds max value ${:.2f}".format(w_max))
        w = min(w, w_max)

        # Recalculate evaporative sweat heat
        p_rsw = (w - wettedness_dif) / (1.0 - wettedness_dif)
        e_rsw = max(0, p_rsw * e_max)
        e_diff = max(0, w * e_max - e_rsw)

        # check condensation on skin surface due to high RH > 100%; body emerged in water;  the model is not valid for this case
        if e_max < 0:
            # sweating is suppressed, and condensation latent heat is not taken into account
            w = 0.0
            e_diff = 0.0
            e_rsw = 0.0

        # Total evaporative heat loss by sweating and vapor diffusion
        e_skin = e_rsw + e_diff

        # Calculate Shivering (SHIV)
        t_cr_sh = max(0, t_cr0_sh - t_core)
        met_shivering = c_she * (
            cof_scs * t_cr_sh * t_sk_cons + cof_sc * t_cr_sh + cof_ss * t_sk_cons
        )
        m = met * met_factor + met_shivering

        # Append skin and core temp for time point
        skin_temp_hist.append(t_skin)
        core_temp_hist.append(t_core)

    return {"t_core": np.array(core_temp_hist), "t_skin": np.array(skin_temp_hist)}



import pytest

from pythermalcomfort.models import pmv_a
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_a_pmv(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the aPMV correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.A_PMV.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = pmv_a(**inputs)

        validate_result(result, outputs, tolerance)


def test_a_pmv_wrong_input_type() -> None:
    """Test that the function raises a TypeError for wrong input types."""
    with pytest.raises(TypeError):
        pmv_a("25", 25, 0.1, 50, 1.2, 0.5, 7)
    with pytest.raises(ValueError):
        pmv_a(25, 25, 0.1, 50, 1.2, 0.5, 7, units="celsius")


def test_not_valid_units() -> None:
    """Test that the function raises a ValueError for invalid units."""
    with pytest.raises(ValueError):
        pmv_a(25, 25, 0.1, 50, 1.2, 0.5, 7, units="wrong")



import numpy as np

from pythermalcomfort.models import utci
from pythermalcomfort.models.utci import _utci_optimized
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_utci(get_test_url, retrieve_data) -> None:
    """Test that the UTCI function calculates correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.UTCI.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = utci(**inputs)

        validate_result(result, outputs, tolerance)


def test_utci_optimized() -> None:
    """Test that the optimized UTCI function calculates correctly for various inputs."""
    np.testing.assert_equal(
        np.around(_utci_optimized([25, 27], 1, 1, 1.5), 2),
        [24.73, 26.57],
    )



from typing import Literal

import numpy as np

from pythermalcomfort.classes_input import ENInputs
from pythermalcomfort.classes_return import AdaptiveEN
from pythermalcomfort.shared_functions import valid_range
from pythermalcomfort.utilities import Units, operative_tmp, units_converter


def adaptive_en(
    tdb: float | list[float],
    tr: float | list[float],
    t_running_mean: float | list[float],
    v: float | list[float],
    units: Literal["SI", "IP"] = Units.SI.value,
    limit_inputs: bool = True,
) -> AdaptiveEN:
    """Calculate the adaptive thermal comfort based on EN 16798-1 2019 [16798EN2019]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, default in [°C] or [°F] if `units` = 'IP'.
    tr : float or list of floats
        Mean radiant temperature, default in [°C] or [°F] if `units` = 'IP'.
    t_running_mean: float or list of floats
        Running mean temperature, default in [°C] or [°F] if `units` = 'IP'.

        .. note::
            The running mean temperature can be calculated using the function :py:meth:`pythermalcomfort.utilities.running_mean_outdoor_temperature`.

    v : float or list of floats
        Air speed, default in [m/s] or [fps] if `units` = 'IP'.

        .. note::
            Indoor operative temperature correction is applicable for buildings equipped
            with fans or personal systems providing building occupants with personal
            control over air speed at occupant level.
            For operative temperatures above 25°C the comfort zone upper limit can be
            increased by 1.2 °C (0.6 < v < 0.9 m/s), 1.8 °C (0.9 < v < 1.2 m/s), 2.2 °C (v
            > 1.2 m/s).

    units : {'SI', 'IP'}
        Select the SI (International System of Units) or the IP (Imperial Units) system.
    limit_inputs : bool, default True
        If True, returns NaN for inputs outside the standard applicability limits.

    Returns
    -------
    AdaptiveEN
        A dataclass containing the results. See :py:class:`~pythermalcomfort.classes_return.AdaptiveEN` for more details.


    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import adaptive_en

        results = adaptive_en(tdb=25, tr=25, t_running_mean=20, v=0.1)
        print(results)
        # AdaptiveEN(tmp_cmf=np.float64(25.4), acceptability_cat_i=np.True_, acceptability_cat_ii=np.True_, ...)

        print(results.acceptability_cat_i)  # or print(results["acceptability_cat_i"])
        # True
        # The conditions you entered are considered to comply with Category I

        # For users who want to use the IP system, units="IP" or "ip" are both valid
        results = adaptive_en(tdb=77, tr=77, t_running_mean=68, v=0.3, units="IP")
        print(results)
        # AdaptiveEN(tmp_cmf=np.float64(77.7), acceptability_cat_i=np.True_, ...)

        results = adaptive_en(tdb=25, tr=25, t_running_mean=9, v=0.1)
        print(results)
        # AdaptiveEN(tmp_cmf=np.float64(nan), acceptability_cat_i=np.False_, ...)
        # The adaptive thermal comfort model can only be used
        # if the running mean temperature is between 10 °C and 30 °C.

    """
    # Validate inputs using the ENInputs class
    ENInputs(tdb=tdb, tr=tr, t_running_mean=t_running_mean, v=v, units=units)

    tdb = np.array(tdb)
    tr = np.array(tr)
    t_running_mean = np.array(t_running_mean)
    v = np.array(v)
    standard = "iso"

    if units.upper() == Units.IP.value:
        tdb, tr, t_running_mean, v = units_converter(
            tdb=tdb,
            tr=tr,
            tmp_running_mean=t_running_mean,
            v=v,
        )

    trm_valid = valid_range(t_running_mean, (10.0, 33.5))

    to = operative_tmp(tdb, tr, v, standard=standard)

    # Calculate cooling effect (ce) of elevated air speed when top > 25 degC.
    ce = np.where((v >= 0.6) & (to >= 25.0), 999, 0)
    ce = np.where((v < 0.9) & (ce == 999), 1.2, ce)
    ce = np.where((v < 1.2) & (ce == 999), 1.8, ce)
    ce = np.where(ce == 999, 2.2, ce)

    t_cmf = 0.33 * t_running_mean + 18.8

    if limit_inputs:
        all_valid = ~(np.isnan(trm_valid))
        t_cmf = np.where(all_valid, t_cmf, np.nan)

    t_cmf_i_lower = t_cmf - 3.0
    t_cmf_ii_lower = t_cmf - 4.0
    t_cmf_iii_lower = t_cmf - 5.0
    t_cmf_i_upper = t_cmf + 2.0 + ce
    t_cmf_ii_upper = t_cmf + 3.0 + ce
    t_cmf_iii_upper = t_cmf + 4.0 + ce

    acceptability_i = (t_cmf_i_lower <= to) & (to <= t_cmf_i_upper)
    acceptability_ii = (t_cmf_ii_lower <= to) & (to <= t_cmf_ii_upper)
    acceptability_iii = (t_cmf_iii_lower <= to) & (to <= t_cmf_iii_upper)

    if units.upper() == Units.IP.value:
        t_cmf, t_cmf_i_upper, t_cmf_ii_upper, t_cmf_iii_upper = units_converter(
            from_units=Units.SI.value.lower(),
            tmp_cmf=t_cmf,
            tmp_cmf_cat_i_up=t_cmf_i_upper,
            tmp_cmf_cat_ii_up=t_cmf_ii_upper,
            tmp_cmf_cat_iii_up=t_cmf_iii_upper,
        )
        t_cmf_i_lower, t_cmf_ii_lower, t_cmf_iii_lower = units_converter(
            from_units=Units.SI.value.lower(),
            tmp_cmf_cat_i_low=t_cmf_i_lower,
            tmp_cmf_cat_ii_low=t_cmf_ii_lower,
            tmp_cmf_cat_iii_low=t_cmf_iii_lower,
        )

    return AdaptiveEN(
        tmp_cmf=np.around(t_cmf, 1),
        acceptability_cat_i=acceptability_i,
        acceptability_cat_ii=acceptability_ii,
        acceptability_cat_iii=acceptability_iii,
        tmp_cmf_cat_i_up=np.around(t_cmf_i_upper, 1),
        tmp_cmf_cat_ii_up=np.around(t_cmf_ii_upper, 1),
        tmp_cmf_cat_iii_up=np.around(t_cmf_iii_upper, 1),
        tmp_cmf_cat_i_low=np.around(t_cmf_i_lower, 1),
        tmp_cmf_cat_ii_low=np.around(t_cmf_ii_lower, 1),
        tmp_cmf_cat_iii_low=np.around(t_cmf_iii_lower, 1),
    )



import pytest

from pythermalcomfort.models import adaptive_en
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_adaptive_en(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the Adaptive Thermal Comfort Model (ASHRAE 55) correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.ADAPTIVE_EN.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = adaptive_en(
            inputs["tdb"],
            inputs["tr"],
            inputs["t_running_mean"],
            inputs["v"],
        )

        validate_result(result, outputs, tolerance)


def test_ashrae_inputs_invalid_units() -> None:
    """Test that the function raises a ValueError for invalid units."""
    with pytest.raises(ValueError):
        adaptive_en(tdb=25, tr=25, t_running_mean=20, v=0.1, units="INVALID")


def test_ashrae_inputs_invalid_tdb_type() -> None:
    """Test that the function raises a TypeError for invalid tdb type."""
    with pytest.raises(TypeError):
        adaptive_en(tdb="invalid", tr=25, t_running_mean=20, v=0.1)


def test_ashrae_inputs_invalid_tr_type() -> None:
    """Test that the function raises a TypeError for invalid tr type."""
    with pytest.raises(TypeError):
        adaptive_en(tdb=25, tr="invalid", t_running_mean=20, v=0.1)


def test_ashrae_inputs_invalid_t_running_mean_type() -> None:
    """Test that the function raises a TypeError for invalid t_running_mean type."""
    with pytest.raises(TypeError):
        adaptive_en(tdb=25, tr=25, t_running_mean="invalid", v=0.1)


def test_ashrae_inputs_invalid_v_type() -> None:
    """Test that the function raises a TypeError for invalid v type."""
    with pytest.raises(TypeError):
        adaptive_en(tdb=25, tr=25, t_running_mean=20, v="invalid")



import pytest

from pythermalcomfort.models import esi
from tests.conftest import is_equal


def test_esi() -> None:
    """Test that the function calculates the ESI correctly for given inputs."""
    result = esi(tdb=30.2, rh=42.2, sol_radiation_global=766)
    is_equal(result.esi, 26.2, 0.1)


def test_esi_list_input() -> None:
    """Test that the function calculates the ESI correctly for list inputs."""
    result = esi([30.2, 27.0], [42.2, 68.8], [766, 289])
    is_equal(result.esi, [26.2, 25.6], 0.1)


@pytest.mark.parametrize(
    ("tdb", "rh", "sol"),
    [
        (30, -5, 500),  # negative relative humidity
        (30, 120, 500),  # RH above 100%
        (30, 50, -10),  # negative solar radiation
    ],
)
def test_esi_invalid_numeric_ranges(tdb, rh, sol) -> None:
    """Test that the function raises ValueError for invalid input ranges."""
    with pytest.raises(ValueError):
        esi(tdb=tdb, rh=rh, sol_radiation_global=sol)


@pytest.mark.parametrize(
    ("tdb", "rh", "sol", "expected"),
    [
        (30, 0, 500, 16.5),  # minimum RH
        (30, 100, 500, 19.5),  # maximum RH
        (30, 50, 0, 18.0),  # minimum solar radiation
    ],
)
def test_esi_boundary_conditions(
    tdb: float,
    rh: float,
    sol: float,
    expected: float,
) -> None:
    """Test that the function handles boundary conditions correctly."""
    result = esi(tdb=tdb, rh=rh, sol_radiation_global=sol)
    is_equal(result.esi, expected, 0.1)  # Replace expected values with actual results


@pytest.mark.parametrize(
    ("tdb", "rh", "sol"),
    [("30.0", 45, 500), (30, "45", 500), (30, 45, "500.0")],
)
def test_esi_invalid_type(tdb: float, rh: float, sol: float) -> None:
    """Test that the function raises TypeError for invalid input types."""
    with pytest.raises(TypeError):
        esi(tdb=tdb, rh=rh, sol_radiation_global=sol)



import itertools

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from pythermalcomfort.models import phs

plt.close("all")

t_array = range(30, 52, 1)
rh_array = range(0, 100, 1)

df = pd.DataFrame(list(itertools.product(t_array, rh_array)), columns=["tdb", "rh"])

results = phs(
    tdb=df.tdb.values,
    tr=df.tdb.values,
    rh=df.rh.values,
    v=0.3,
    met=2,
    clo=0.5,
    posture="sitting",
)

df["t_re"] = results.t_re

plt.figure()
pivot = df.pivot(columns="tdb", index="rh", values="t_re")
pivot = pivot.sort_index(ascending=False)
ax = sns.heatmap(pivot)
plt.tight_layout()
plt.show()



from typing import Literal

import numpy as np

from pythermalcomfort.classes_input import ASHRAEInputs
from pythermalcomfort.classes_return import AdaptiveASHRAE
from pythermalcomfort.shared_functions import valid_range
from pythermalcomfort.utilities import (
    Models,
    Units,
    _check_standard_compliance_array,
    operative_tmp,
    units_converter,
)


def adaptive_ashrae(
    tdb: float | list[float],
    tr: float | list[float],
    t_running_mean: float | list[float],
    v: float | list[float],
    units: Literal["SI", "IP"] = Units.SI.value,
    limit_inputs: bool = True,
) -> AdaptiveASHRAE:
    """Calculate the adaptive thermal comfort based on ASHRAE 55.

    The adaptive
    model relates indoor design temperatures or acceptable temperature ranges
    to outdoor meteorological or climatological parameters. The adaptive model
    can only be used in occupant-controlled naturally conditioned spaces that
    meet all the following criteria:

    * There is no mechanical cooling or heating system in operation.
    * Occupants have a metabolic rate between 1.0 and 1.5 met.
    * Occupants are free to adapt their clothing within a range as wide as 0.5 and 1.0 clo.
    * The prevailing mean (running mean) outdoor temperature is between 10 and 33.5 °C.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, default in [°C] or [°F] if `units` = 'IP'.
    tr : float or list of floats
        Mean radiant temperature, default in [°C] or [°F] if `units` = 'IP'.
    t_running_mean : float or list of floats
        Running mean temperature, default in [°C] or [°F] if `units` = 'IP'.

        .. note::
            The running mean temperature can be calculated using the function :py:meth:`pythermalcomfort.utilities.running_mean_outdoor_temperature`.

    v : float or list of floats
        Air speed, default in [m/s] or [fps] if `units` = 'IP'.
    units : str, optional
        Units system, 'SI' or 'IP'. Defaults to 'SI'.
    limit_inputs : bool, optional
        If True, returns nan for inputs outside standard limits. Defaults to True.

        .. note::
            ASHRAE 55 2020 limits: 10 < tdb [°C] < 40, 10 < tr [°C] < 40, 0 < vr [m/s] < 2, 10 < t_running_mean [°C] < 33.5.

    Returns
    -------
    AdaptiveASHRAE
        A dataclass containing the results. See :py:class:`~pythermalcomfort.classes_return.AdaptiveASHRAE` for more details.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import adaptive_ashrae

        results = adaptive_ashrae(tdb=25, tr=25, t_running_mean=20, v=0.1)
        print(results)
        # AdaptiveASHRAE(tmp_cmf=np.float64(24.0), tmp_cmf_80_low=np.float64(20.5), tmp_cmf_80_up=np.float64(27.5), tmp_cmf_90_low=np.float64(21.5), tmp_cmf_90_up=np.float64(26.5), acceptability_80=array(True), acceptability_90=array(True))

        print(results.acceptability_80)  # or use print(results["acceptability_80"])
        # True
        # The conditions you entered are considered to be comfortable for 80% of the occupants.

        # You can also pass arrays as input to the function
        results = adaptive_ashrae(tdb=[25, 26], tr=25, t_running_mean=20, v=0.1)
        print(results)
        # AdaptiveASHRAE(tmp_cmf=array([24., 24.]), tmp_cmf_80_low=array([20.5, 20.5]), tmp_cmf_80_up=array([27.5, 27.5]), tmp_cmf_90_low=array([21.5, 21.5]), tmp_cmf_90_up=array([26.5, 26.5]), acceptability_80=array([ True,  True]), acceptability_90=array([ True,  True]))

        # For users who want to use the IP system
        results = adaptive_ashrae(tdb=77, tr=77, t_running_mean=68, v=0.3, units="IP")
        print(results)
        # AdaptiveASHRAE(tmp_cmf=np.float64(75.2), tmp_cmf_80_low=np.float64(68.9), tmp_cmf_80_up=np.float64(81.5), tmp_cmf_90_low=np.float64(70.7), tmp_cmf_90_up=np.float64(79.7), acceptability_80=array(True), acceptability_90=array(True))

        adaptive_ashrae(tdb=25, tr=25, t_running_mean=9, v=0.1)
        # AdaptiveASHRAE(tmp_cmf=np.float64(nan), ... acceptability_90=array(False))
        # The adaptive thermal comfort model can only be used if the running mean temperature is higher than 10°C.

    """
    # Validate inputs using the ASHRAEInputs class
    ASHRAEInputs(
        tdb=tdb,
        tr=tr,
        t_running_mean=t_running_mean,
        v=v,
        units=units,
    )

    tdb = np.array(tdb)
    tr = np.array(tr)
    t_running_mean = np.array(t_running_mean)
    v = np.array(v)
    standard = "ashrae"

    if units.upper() == Units.IP.value:
        tdb, tr, t_running_mean, v = units_converter(
            tdb=tdb,
            tr=tr,
            tmp_running_mean=t_running_mean,
            v=v,
        )

    tdb_valid, tr_valid, v_valid = _check_standard_compliance_array(
        Models.ashrae_55_2023.value,
        tdb=tdb,
        tr=tr,
        v=v,
    )
    trm_valid = valid_range(t_running_mean, (10.0, 33.5))

    to = operative_tmp(tdb, tr, v, standard=standard)

    # Calculate cooling effect (ce) of elevated air speed when top > 25 degC.
    ce = np.where((v >= 0.6) & (to >= 25.0), 999, 0)
    ce = np.where((v < 0.9) & (ce == 999), 1.2, ce)
    ce = np.where((v < 1.2) & (ce == 999), 1.8, ce)
    ce = np.where(ce == 999, 2.2, ce)

    # Relation between comfort and outdoor temperature
    t_cmf = 0.31 * t_running_mean + 17.8

    if limit_inputs:
        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(trm_valid)
        )
        t_cmf = np.where(all_valid, t_cmf, np.nan)

    t_cmf = np.around(t_cmf, 1)

    tmp_cmf_80_low = t_cmf - 3.5
    tmp_cmf_90_low = t_cmf - 2.5
    tmp_cmf_80_up = t_cmf + 3.5 + ce
    tmp_cmf_90_up = t_cmf + 2.5 + ce

    acceptability_80 = (tmp_cmf_80_low <= to) & (to <= tmp_cmf_80_up)
    acceptability_90 = (tmp_cmf_90_low <= to) & (to <= tmp_cmf_90_up)

    if units.upper() == Units.IP.value:
        (
            t_cmf,
            tmp_cmf_80_low,
            tmp_cmf_80_up,
            tmp_cmf_90_low,
            tmp_cmf_90_up,
        ) = units_converter(
            from_units=Units.SI.value,
            tmp_cmf=t_cmf,
            tmp_cmf_80_low=tmp_cmf_80_low,
            tmp_cmf_80_up=tmp_cmf_80_up,
            tmp_cmf_90_low=tmp_cmf_90_low,
            tmp_cmf_90_up=tmp_cmf_90_up,
        )

    return AdaptiveASHRAE(
        tmp_cmf=t_cmf,
        tmp_cmf_80_low=tmp_cmf_80_low,
        tmp_cmf_80_up=tmp_cmf_80_up,
        tmp_cmf_90_low=tmp_cmf_90_low,
        tmp_cmf_90_up=tmp_cmf_90_up,
        acceptability_80=acceptability_80,
        acceptability_90=acceptability_90,
    )



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import WCTInputs
from pythermalcomfort.classes_return import WCT


def wind_chill_temperature(
    tdb: float | list[float],
    v: float | list[float],
    round_output: bool = True,
) -> WCT:
    """Calculate the Wind Chill Temperature (`WCT`_).

    We validated the implementation of this model by comparing the results with the Wind Chill
    Temperature Calculator on `Calculator.net`_

    .. _WCT: https://en.wikipedia.org/wiki/Wind_chill#North_American_and_United_Kingdom_wind_chill_index
    .. _Calculator.net: https://www.calculator.net/wind-chill-calculator.html

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    v : float or list of floats
        Wind speed 10m above ground level, [km/h].
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    WCT
        A dataclass containing the Wind Chill Temperature. See :py:class:`~pythermalcomfort.classes_return.WCT` for more details.
        To access the `wct` value, use the `wct` attribute of the returned `WCT` instance, e.g., `result.wct`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import wind_chill_temperature

        result = wind_chill_temperature(tdb=-5, v=5.5)
        print(result.wct)  # -7.5

        result = wind_chill_temperature(tdb=[-5, -10], v=[5.5, 10], round_output=True)
        print(result.wct)  # [-7.5, -15.3]

    """
    # Validate inputs using the WCYInputs class
    WCTInputs(
        tdb=tdb,
        v=v,
        round_output=round_output,
    )

    tdb = np.array(tdb)
    v = np.array(v)

    _wct = 13.12 + 0.6215 * tdb - 11.37 * v**0.16 + 0.3965 * tdb * v**0.16

    if round_output:
        _wct = np.around(_wct, 1)

    return WCT(wct=_wct)



from dataclasses import dataclass

import pytest

# Import the AutoStrMixin class
from pythermalcomfort.classes_return import AutoStrMixin


@dataclass(repr=False)
class TestDataClass(AutoStrMixin):
    """A test dataclass to demonstrate the AutoStrMixin functionality."""

    field1: int
    field2: str
    field3: list


def test_autostr_with_dataclass() -> None:
    """Test that the AutoStrMixin generates a string representation for a dataclass."""
    obj = TestDataClass(field1=42, field2="test", field3=[1, 2, 3])
    expected_output = (
        "-------- TestDataClass --------\n"
        "field1 : 42\n"
        "field2 : test\n"
        "field3 : [1, 2, 3]"
    )
    assert str(obj) == expected_output


def test_autostr_with_multiline_field() -> None:
    """Test that the AutoStrMixin handles multiline fields correctly."""
    obj = TestDataClass(field1=42, field2="test", field3=["line1", "line2"])
    expected_output = (
        "-------- TestDataClass --------\n"
        "field1 : 42\n"
        "field2 : test\n"
        "field3 : ['line1', 'line2']"
    )
    assert str(obj) == expected_output


def test_autostr_empty_dataclass() -> None:
    """Test that the AutoStrMixin handles an empty dataclass correctly."""

    @dataclass
    class EmptyDataClass(AutoStrMixin):
        pass

    obj = EmptyDataClass()
    expected_output = "-------- EmptyDataClass --------"
    assert str(obj) == expected_output


def test_autostr_repr_method() -> None:
    """Test that the __repr__ method returns the same as __str__."""
    obj = TestDataClass(field1=42, field2="test", field3=[1, 2, 3])
    expected_output = (
        "-------- TestDataClass --------\n"
        "field1 : 42\n"
        "field2 : test\n"
        "field3 : [1, 2, 3]"
    )
    assert repr(obj) == expected_output
    # Verify that __repr__ returns the same as __str__
    assert repr(obj) == str(obj)


def test_autostr_getitem_method() -> None:
    """Test that the __getitem__ method works correctly."""
    obj = TestDataClass(field1=42, field2="test", field3=[1, 2, 3])
    assert obj["field1"] == 42
    assert obj["field2"] == "test"
    assert obj["field3"] == [1, 2, 3]


def test_autostr_getitem_method_key_error() -> None:
    """Test that __getitem__ raises AttributeError for non-existent keys."""
    obj = TestDataClass(field1=42, field2="test", field3=[1, 2, 3])
    with pytest.raises(KeyError):
        _ = obj["non_existent"]



from __future__ import annotations

import datetime as dt
from dataclasses import dataclass, fields, is_dataclass

import numpy as np
import numpy.typing as npt


class AutoStrMixin:
    def __str__(self) -> str:
        if not is_dataclass(self):
            return super().__str__()

        # determine width by max variable name length
        names = [f.name for f in fields(self)]
        width = max((len(n) for n in names), default=0)
        lines = [f"-------- {self.__class__.__name__} --------"]
        for n in names:
            v = getattr(self, n)
            # Format multi-line values or very long values properly
            v_str = str(v).replace("\n", "\n" + " " * (width + 3 + 3))
            lines.append(f"{n.ljust(width)} : {v_str}")
        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.__str__()

    def __getitem__(self, item: str):
        try:
            return getattr(self, item)
        except AttributeError as exc:
            error_msg = (
                f"{self.__class__.__name__} has no field '{item}'. "
                f"Available fields: {[f.name for f in fields(self)]}"
            )
            raise KeyError(error_msg) from exc


@dataclass(frozen=True, repr=False)
class APMV(AutoStrMixin):
    """A dataclass to store the results of the adaptive Predicted Mean Vote (aPMV)
    model.

    Attributes
    ----------
    a_pmv : float or list of floats
        Predicted Mean Vote.

    """

    a_pmv: float | list[float]


@dataclass(frozen=True, repr=False)
class AdaptiveASHRAE(AutoStrMixin):
    """A dataclass to store the results of the adaptive thermal comfort model based on
    ASHRAE 55.

    Attributes
    ----------
    tmp_cmf : float or list of floats
        Comfort temperature at a specific running mean temperature, default in [°C] or [°F].
    tmp_cmf_80_low : float or list of floats
        Lower acceptable comfort temperature for 80% occupants, default in [°C] or [°F].
    tmp_cmf_80_up : float or list of floats
        Upper acceptable comfort temperature for 80% occupants, default in [°C] or [°F].
    tmp_cmf_90_low : float or list of floats
        Lower acceptable comfort temperature for 90% occupants, default in [°C] or [°F].
    tmp_cmf_90_up : float or list of floats
        Upper acceptable comfort temperature for 90% occupants, default in [°C] or [°F].
    acceptability_80 : bool or list of bools
        Acceptability for 80% occupants.
    acceptability_90 : bool or list of bools
        Acceptability for 90% occupants.

    """

    tmp_cmf: float | list[float]
    tmp_cmf_80_low: float | list[float]
    tmp_cmf_80_up: float | list[float]
    tmp_cmf_90_low: float | list[float]
    tmp_cmf_90_up: float | list[float]
    acceptability_80: bool | list[bool]
    acceptability_90: bool | list[bool]


@dataclass
class AdaptiveEN(AutoStrMixin):
    """Dataclass to store the results of the adaptive thermal comfort calculation based
    on EN 16798-1 2019.

    Attributes
    ----------
    tmp_cmf : float or list of floats
        Comfort temperature at that specific running mean temperature, default in [°C] or in [°F].
    acceptability_cat_i : bool or list of bools
        If the indoor conditions comply with comfort category I.
    acceptability_cat_ii : bool or list of bools
        If the indoor conditions comply with comfort category II.
    acceptability_cat_iii : bool or list of bools
        If the indoor conditions comply with comfort category III.
    tmp_cmf_cat_i_up : float or list of floats
        Upper acceptable comfort temperature for category I, default in [°C] or in [°F].
    tmp_cmf_cat_ii_up : float or list of floats
        Upper acceptable comfort temperature for category II, default in [°C] or in [°F].
    tmp_cmf_cat_iii_up : float or list of floats
        Upper acceptable comfort temperature for category III, default in [°C] or in [°F].
    tmp_cmf_cat_i_low : float or list of floats
        Lower acceptable comfort temperature for category I, default in [°C] or in [°F].
    tmp_cmf_cat_ii_low : float or list of floats
        Lower acceptable comfort temperature for category II, default in [°C] or in [°F].
    tmp_cmf_cat_iii_low : float or list of floats
        Lower acceptable comfort temperature for category III, default in [°C] or in [°F].

    """

    tmp_cmf: float | list[float]
    acceptability_cat_i: bool | list[bool]
    acceptability_cat_ii: bool | list[bool]
    acceptability_cat_iii: bool | list[bool]
    tmp_cmf_cat_i_up: float | list[float]
    tmp_cmf_cat_ii_up: float | list[float]
    tmp_cmf_cat_iii_up: float | list[float]
    tmp_cmf_cat_i_low: float | list[float]
    tmp_cmf_cat_ii_low: float | list[float]
    tmp_cmf_cat_iii_low: float | list[float]


@dataclass(frozen=True, repr=False)
class AnkleDraft(AutoStrMixin):
    """Dataclass to store the results of the ankle draft calculation.

    Attributes
    ----------
    ppd_ad : float or list of floats
        Predicted Percentage of Dissatisfied occupants with ankle draft, [%].
    acceptability : bool or list of bools
        Indicates if the air speed at the ankle level is acceptable according to ASHRAE 55 2020 standard.

    """

    ppd_ad: float | list[float]
    acceptability: bool | list[bool]


@dataclass(frozen=True, repr=False)
class AT(AutoStrMixin):
    """Dataclass to store the results of the Apparent Temperature (AT) calculation.

    Attributes
    ----------
    at : float or list of floats
        Apparent temperature, [°C]

    """

    at: float


@dataclass(frozen=True, repr=False)
class ATHB(AutoStrMixin):
    """Dataclass to store the results of the Adaptive Thermal Heat Balance (ATHB)
    calculation.

    Attributes
    ----------
    athb_pmv : float or list of floats
        Predicted Mean Vote calculated with the Adaptive Thermal Heat Balance framework.

    """

    athb_pmv: float | list[float]


@dataclass(frozen=True, repr=False)
class CloTOut(AutoStrMixin):
    """Dataclass to represent the clothing insulation Icl as a function of outdoor air
    temperature.

    Attributes
    ----------
    clo_tout : float or list of floats
        Representative clothing insulation Icl.

    """

    clo_tout: float | list[float]


@dataclass(frozen=True, repr=False)
class CE(AutoStrMixin):
    """Dataclass to represent the Cooling Effect (CE).

    Attributes
    ----------
    ce : float or list of floats
        Cooling Effect value.

    """

    ce: float | list[float]


@dataclass(frozen=True, repr=False)
class DI(AutoStrMixin):
    """Dataclass to represent the Discomfort Index (DI) and its classification.

    Attributes
    ----------
    di : float or list of floats
        Discomfort Index, [°C].
    discomfort_condition : str or list of str
        Classification of the thermal comfort conditions according to the discomfort index.

    """

    di: float | list[float]
    discomfort_condition: str | list[str]


@dataclass(frozen=True, repr=False)
class EPMV(AutoStrMixin):
    """Dataclass to represent the Adjusted Predicted Mean Votes with Expectancy Factor
    (ePMV).

    Attributes
    ----------
    e_pmv : float or list of floats
        Adjusted Predicted Mean Votes with Expectancy Factor.

    """

    e_pmv: float | list[float]


@dataclass(frozen=True, repr=False)
class ESI(AutoStrMixin):
    """Dataclass to represent the Environmental Stress Index (ESI).

    Attributes
    ----------
    esi : float or list of floats
        Environmental Stress Index.

    """

    esi: float | list[float]


@dataclass(frozen=True, repr=False)
class HI(AutoStrMixin):
    """Dataclass to represent the Heat Index (HI).

    Attributes
    ----------
    hi : float or list of floats
        Heat Index, [°C] or [°F] depending on the units.

    """

    hi: npt.ArrayLike
    stress_category: str | list[str] | None = None


@dataclass(frozen=True, repr=False)
class Humidex(AutoStrMixin):
    """Dataclass to represent the Humidex and its discomfort category.

    Attributes
    ----------
    humidex : float or list of floats
        Humidex value, [°C].
    discomfort : str or list of str
        Degree of comfort or discomfort as defined in Havenith and Fiala (2016).

    """

    humidex: float | list[float]
    discomfort: str | list[str]


@dataclass(frozen=True, repr=False)
class NET(AutoStrMixin):
    """Dataclass to represent the Normal Effective Temperature (NET).

    Attributes
    ----------
    net : float or list of floats
        Normal Effective Temperature, [°C].

    """

    net: float | list[float]


@dataclass(frozen=True, repr=False)
class PETSteady(AutoStrMixin):
    """Dataclass to represent the Physiological Equivalent Temperature (PET).

    Attributes
    ----------
    pet : float or list of floats
        Physiological Equivalent Temperature.

    """

    pet: float | list[float]


@dataclass(frozen=True, repr=False)
class PHS(AutoStrMixin):
    """Dataclass to represent the Predicted Heat Strain (PHS).

    Attributes
    ----------
    t_re : float or list of floats
        Rectal temperature, [°C].
    t_sk : float or list of floats
        Skin temperature, [°C].
    t_cr : float or list of floats
        Core temperature, [°C].
    t_cr_eq : float or list of floats
        Core temperature as a function of the metabolic rate, [°C].
    t_sk_t_cr_wg : float or list of floats
        Fraction of the body mass at the skin temperature.
    d_lim_loss_50 : float or list of floats
        Maximum allowable exposure time for water loss, mean subject, [minutes].
    d_lim_loss_95 : float or list of floats
        Maximum allowable exposure time for water loss, 95% of the working population, [minutes].
    d_lim_t_re : float or list of floats
        Maximum allowable exposure time for heat storage, [minutes].
    water_loss_watt : float or list of floats
        Maximum water loss in watts, [W].
    water_loss : float or list of floats
        Maximum water loss, [g].

    """

    t_re: float | list[float]
    t_sk: float | list[float]
    t_cr: float | list[float]
    t_cr_eq: float | list[float]
    t_sk_t_cr_wg: float | list[float]
    d_lim_loss_50: float | list[float]
    d_lim_loss_95: float | list[float]
    d_lim_t_re: float | list[float]
    water_loss_watt: float | list[float]
    water_loss: float | list[float]


@dataclass(frozen=True, repr=False)
class PMV(AutoStrMixin):
    """Dataclass to represent the Predicted Mean Vote (PMV).

    Attributes
    ----------
    pmv : float or list of floats
        Predicted Mean Vote.

    """

    pmv: float | list[float]


@dataclass(frozen=True, repr=False)
class PMVPPD(AutoStrMixin):
    """Dataclass to represent the Predicted Mean Vote (PMV) and Predicted Percentage of
    Dissatisfied (PPD).

    Attributes
    ----------
    pmv : float or list of floats
        Predicted Mean Vote.
    ppd : float or list of floats
        Predicted Percentage of Dissatisfied.
    tsv : str or list of strings
        Predicted thermal sensation vote.

    """

    pmv: float | list[float]
    ppd: float | list[float]
    tsv: float | list[float]


@dataclass(frozen=True, repr=False)
class PsychrometricValues(AutoStrMixin):
    p_sat: float | list[float]
    p_vap: float | list[float]
    hr: float | list[float]
    wet_bulb_tmp: float | list[float]
    dew_point_tmp: float | list[float]
    h: float | list[float]


@dataclass(frozen=True, repr=False)
class SET(AutoStrMixin):
    """Dataclass to represent the Standard Effective Temperature (SET).

    Attributes
    ----------
    set : float or list of floats
        Standard effective temperature, [°C].

    """

    set: float | list[float]


@dataclass(frozen=True, repr=False)
class SolarGain(AutoStrMixin):
    """Dataclass to represent the solar gain to the human body.

    Attributes
    ----------
    erf : float or list of floats
        Solar gain to the human body using the Effective Radiant Field [W/m2].
    delta_mrt : float or list of floats
        Delta mean radiant temperature. The amount by which the mean radiant
        temperature of the space should be increased if no solar radiation is present.

    """

    erf: float | list[float]
    delta_mrt: float | list[float]


@dataclass(frozen=True, repr=False)
class GaggeTwoNodes(AutoStrMixin):
    """Dataclass to represent the results of the two-node model of human temperature
    regulation.

    Attributes
    ----------
    e_skin : float or list of floats
        Total rate of evaporative heat loss from skin, [W/m2]. Equal to e_rsw + e_diff.
    e_rsw : float or list of floats
        Rate of evaporative heat loss from sweat evaporation, [W/m2].
    e_max : float or list of floats
        Maximum rate of evaporative heat loss from skin, [W/m2].
    q_sensible : float or list of floats
        Sensible heat loss from skin, [W/m2].
    q_skin : float or list of floats
        Total rate of heat loss from skin, [W/m2]. Equal to q_sensible + e_skin.
    q_res : float or list of floats
        Total rate of heat loss through respiration, [W/m2].
    t_core : float or list of floats
        Core temperature, [°C].
    t_skin : float or list of floats
        Skin temperature, [°C].
    m_bl : float or list of floats
        Skin blood flow, [kg/h/m2].
    m_rsw : float or list of floats
        Rate at which regulatory sweat is generated, [mL/h/m2].
    w : float or list of floats
        Skin wettedness, adimensional. Ranges from 0 to 1.
    w_max : float or list of floats
        Skin wettedness (w) practical upper limit, adimensional. Ranges from 0 to 1.
    set : float or list of floats
        Standard Effective Temperature (SET).
    et : float or list of floats
        New Effective Temperature (ET).
    pmv_gagge : float or list of floats
        PMV Gagge.
    pmv_set : float or list of floats
        PMV SET.
    disc : float or list of floats
        Thermal discomfort.
    t_sens : float or list of floats
        Predicted Thermal Sensation.

    """

    e_skin: float | list[float]
    e_rsw: float | list[float]
    e_max: float | list[float]
    q_sensible: float | list[float]
    q_skin: float | list[float]
    q_res: float | list[float]
    t_core: float | list[float]
    t_skin: float | list[float]
    m_bl: float | list[float]
    m_rsw: float | list[float]
    w: float | list[float]
    w_max: float | list[float]
    set: float | list[float]
    et: float | list[float]
    pmv_gagge: float | list[float]
    pmv_set: float | list[float]
    disc: float | list[float]
    t_sens: float | list[float]


@dataclass(frozen=True, repr=False)
class GaggeTwoNodesJi(AutoStrMixin):
    """Dataclass to represent the results of the Gagge-Ji model of human temperature.

    Attributes
    ----------
    t_core : float or list of floats
        Core temperature, [°C].
    t_skin : float or list of floats
        Skin temperature, [°C].

    """

    t_core: float | list[float]
    t_skin: float | list[float]


@dataclass(frozen=True, repr=False)
class THI(AutoStrMixin):
    """Dataclass to represent the Temperature-Humidity Index (THI).

    Attributes
    ----------
    thi : float or list of floats
        Temperature-Humidity Index (THI).

    """

    thi: float | list[float]


@dataclass(frozen=True, repr=False)
class GaggeTwoNodesSleep(AutoStrMixin):
    """Dataclass to represent the results of the two-node sleep model.

    Attributes
    ----------
    set : float or list of floats
        Standard Effective Temperature (SET).
    t_core : float or list of floats
        Core temperature, [°C].
    t_skin : float or list of floats
        Skin temperature, [°C].
    wet : float or list of floats
        Skin wettedness, adimensional. Ranges from 0 to 1.
    t_sens : float or list of floats
        Predicted Thermal Sensation.
    disc : float or list of floats
        Thermal discomfort.
    e_skin : float or list of floats
        Total rate of evaporative heat loss from skin, [W/m2]. Equal to e_rsw + e_diff.
    met_shivering : float or list of floats
        Metabolic rate due to shivering, [W/m2].
    alfa : float or list of floats
        Dynamic fraction of total body mass assigned to the skin node (dimensionless).
    skin_blood_flow: float or list of floats
        Skin-blood-flow rate per unit surface area, [kg/h/m2].

    """

    set: float | list[float]
    t_core: float | list[float]
    t_skin: float | list[float]
    wet: float | list[float]
    t_sens: float | list[float]
    disc: float | list[float]
    e_skin: float | list[float]
    met_shivering: float | list[float]
    alfa: float | list[float]
    skin_blood_flow: float | list[float]


@dataclass(frozen=True)
class UseFansHeatwaves(AutoStrMixin):
    """Dataclass to represent the results of using fans during heatwaves.

    Attributes
    ----------
    e_skin : float or list of floats
        Total rate of evaporative heat loss from skin, [W/m2]. Equal to e_rsw + e_diff.
    e_rsw : float or list of floats
        Rate of evaporative heat loss from sweat evaporation, [W/m2].
    e_max : float or list of floats
        Maximum rate of evaporative heat loss from skin, [W/m2].
    q_sensible : float or list of floats
        Sensible heat loss from skin, [W/m2].
    q_skin : float or list of floats
        Total rate of heat loss from skin, [W/m2]. Equal to q_sensible + e_skin.
    q_res : float or list of floats
        Total rate of heat loss through respiration, [W/m2].
    t_core : float or list of floats
        Core temperature, [°C].
    t_skin : float or list of floats
        Skin temperature, [°C].
    m_bl : float or list of floats
        Skin blood flow, [kg/h/m2].
    m_rsw : float or list of floats
        Rate at which regulatory sweat is generated, [mL/h/m2].
    w : float or list of floats
        Skin wettedness, adimensional. Ranges from 0 to 1.
    w_max : float or list of floats
        Skin wettedness (w) practical upper limit, adimensional. Ranges from 0 to 1.
    heat_strain : bool or list of bools
        True if the model predicts that the person may be experiencing heat strain.
    heat_strain_blood_flow : bool or list of bools
        True if heat strain is caused by skin blood flow (m_bl) reaching its maximum value.
    heat_strain_w : bool or list of bools
        True if heat strain is caused by skin wettedness (w) reaching its maximum value.
    heat_strain_sweating : bool or list of bools
        True if heat strain is caused by regulatory sweating (m_rsw) reaching its maximum value.

    """

    e_skin: float | list[float]
    e_rsw: float | list[float]
    e_max: float | list[float]
    q_sensible: float | list[float]
    q_skin: float | list[float]
    q_res: float | list[float]
    t_core: float | list[float]
    t_skin: float | list[float]
    m_bl: float | list[float]
    m_rsw: float | list[float]
    w: float | list[float]
    w_max: float | list[float]
    heat_strain: bool | list[bool]
    heat_strain_blood_flow: bool | list[bool]
    heat_strain_w: bool | list[bool]
    heat_strain_sweating: bool | list[bool]


@dataclass(frozen=True, repr=False)
class UTCI(AutoStrMixin):
    """Dataclass to represent the Universal Thermal Climate Index (UTCI).

    Attributes
    ----------
    utci : float or list of floats
        Universal Thermal Climate Index, [°C] or in [°F].
    stress_category : str or list of strs
        UTCI categorized in terms of thermal stress [Blazejczyk2013]_.

    """

    utci: float | list[float]
    stress_category: str | list[str]


@dataclass(frozen=True, repr=False)
class VerticalTGradPPD(AutoStrMixin):
    """Dataclass to represent the Predicted Percentage of Dissatisfied (PPD) with
    vertical temperature gradient.

    Attributes
    ----------
    ppd_vg : float or list of floats
        Predicted Percentage of Dissatisfied occupants with vertical temperature gradient.
    acceptability : bool or list of bools
        True if the value of air speed at the ankle level is acceptable (PPD_vg <= 5%).

    """

    ppd_vg: float | list[float]
    acceptability: bool | list[bool]


@dataclass(frozen=True, repr=False)
class WBGT(AutoStrMixin):
    """Dataclass to represent the Wet Bulb Globe Temperature (WBGT) index.

    Attributes
    ----------
    wbgt : float or list of floats
        Wet Bulb Globe Temperature Index.

    """

    wbgt: float | list[float]


@dataclass(frozen=True, repr=False)
class WCI(AutoStrMixin):
    """Dataclass to represent the Wind Chill Index (WCI).

    Attributes
    ----------
    wci : float or list of floats
        Wind Chill Index, [W/m^2].

    """

    wci: float | list[float]


@dataclass(frozen=True, repr=False)
class WCT(AutoStrMixin):
    """Dataclass to represent the Wind Chill Temperature (WCT).

    Attributes
    ----------
    wct : float or list of floats
        Wind Chill Temperature, [°C].

    """

    wct: float | list[float]


@dataclass(frozen=True)
class WorkCapacity(AutoStrMixin):
    """Dataclass to represent work loss.

    Attributes
    ----------
    capacity : float or list of floats
        Work capacity affected by heat.

    """

    capacity: float | list[float]


@dataclass(frozen=True, repr=False)
class JOS3BodyParts(AutoStrMixin):
    """Dataclass to represent the body parts in the JOS3 model.

    It is very important to
    keep the order of the attributes as they are defined in the dataclass ['head',
    'neck', 'chest', 'back', 'pelvis', 'left_shoulder', 'left_arm', 'left_hand',
    'right_shoulder', 'right_arm', 'right_hand', 'left_thigh', 'left_leg', 'left_foot',
    'right_thigh', 'right_leg', 'right_foot']

    Attributes
    ----------
    head : float
        Index of the head.
    neck : float
        Index of the neck.
    chest : float
        Index of the chest.
    back : float
        Index of the back.
    pelvis : float
        Index of the pelvis.
    left_shoulder : float
        Index of the left shoulder.
    left_arm : float
        Index of the left arm.
    left_hand : float
        Index of the left hand.
    right_shoulder : float
        Index of the right shoulder.
    right_arm : float
        Index of the right arm.
    right_hand : float
        Index of the right hand.
    left_thigh : float
        Index of the left thigh.
    left_leg : float
        Index of the left leg.
    left_foot : float
        Index of the left foot.
    right_thigh : float
        Index of the right thigh.
    right_leg : float
        Index of the right leg.
    right_foot : float
        Index of the right hand.

    """

    head: float | None = None
    neck: float | None = None
    chest: float | None = None
    back: float | None = None
    pelvis: float | None = None
    left_shoulder: float | None = None
    left_arm: float | None = None
    left_hand: float | None = None
    right_shoulder: float | None = None
    right_arm: float | None = None
    right_hand: float | None = None
    left_thigh: float | None = None
    left_leg: float | None = None
    left_foot: float | None = None
    right_thigh: float | None = None
    right_leg: float | None = None
    right_foot: float | None = None

    @classmethod
    def get_attribute_names(cls):
        return [field.name for field in fields(cls)]


def get_attribute_values(cls):
    return np.array([getattr(cls, field.name) for field in fields(cls)])


@dataclass(frozen=True, repr=False)
class JOS3Output(AutoStrMixin):
    """Dataclass to represent the output of the JOS3 model simulation.

    Attributes
    ----------
    simulation_time : datetime.timedelta
        The elapsed simulation time.
    dt : int or float
        The time step in seconds.
    t_skin_mean : float
        Mean skin temperature of the whole body [°C].
    t_skin : np.ndarray
        Skin temperatures by the local body segments [°C].
    t_core : np.ndarray
        Core temperatures by the local body segments [°C].
    w_mean : float
        Mean skin wettedness of the whole body [-].
    w : np.ndarray
        Skin wettedness on local body segments [-].
    weight_loss_by_evap_and_res : float
        Weight loss by evaporation and respiration [g/sec].
    cardiac_output : float
        Cardiac output [L/h].
    q_thermogenesis_total : float
        Total thermogenesis [W].
    q_res : float
        Heat loss by respiration [W].
    q_skin2env : np.ndarray
        Total heat loss from the skin to the environment [W].
    height : float
        Body height [m].
    weight : float
        Body weight [kg].
    bsa : np.ndarray
        Body surface area.
    fat : float
        Body fat rate [%].
    sex : str
        Sex.
    age : int
        Age [years].
    t_core_set : np.ndarray
        Core set point temperature [°C].
    t_skin_set : np.ndarray
        Skin set point temperature [°C].
    t_cb : float
        Central blood temperature [°C].
    t_artery : np.ndarray
        Arterial blood temperature [°C].
    t_vein : np.ndarray
        Venous blood temperature [°C].
    t_superficial_vein : np.ndarray
        Superficial venous blood temperature [°C].
    t_muscle : np.ndarray
        Muscle temperature [°C].
    t_fat : np.ndarray
        Fat temperature [°C].
    to : float
        Operative temperature [°C].
    r_t : np.ndarray
        Radiative heat transfer coefficient.
    r_et : np.ndarray
        Evaporative heat transfer coefficient.
    tdb : np.ndarray
        Dry bulb air temperature [°C].
    tr : np.ndarray
        Mean radiant temperature [°C].
    rh : np.ndarray
        Relative humidity [%].
    v : np.ndarray
        Air velocity [m/s].
    par : float
        Physical activity ratio.
    clo : np.ndarray
        Clothing insulation.
    e_skin : np.ndarray
        Evaporative heat loss from the skin [W].
    e_max : np.ndarray
        Maximum evaporative heat loss from the skin [W].
    e_sweat : np.ndarray
        Evaporative heat loss from the skin by only sweating [W].
    bf_core : np.ndarray
        Core blood flow rate [L/h].
    bf_muscle : np.ndarray
        Muscle blood flow rate [L/h].
    bf_fat : np.ndarray
        Fat blood flow rate [L/h].
    bf_skin : np.ndarray
        Skin blood flow rate [L/h].
    bf_ava_hand : np.ndarray
        AVA blood flow rate of one hand [L/h].
    bf_ava_foot : np.ndarray
        AVA blood flow rate of one foot [L/h].
    q_bmr_core : np.ndarray
        Core thermogenesis by basal metabolism [W].
    q_bmr_muscle : np.ndarray
        Muscle thermogenesis by basal metabolism [W].
    q_bmr_fat : np.ndarray
        Fat thermogenesis by basal metabolism [W].
    q_bmr_skin : np.ndarray
        Skin thermogenesis by basal metabolism [W].
    q_work : np.ndarray
        Thermogenesis by work [W].
    q_shiv : np.ndarray
        Thermogenesis by shivering [W].
    q_nst : np.ndarray
        Thermogenesis by non-shivering [W].
    q_thermogenesis_core : np.ndarray
        Core total thermogenesis [W].
    q_thermogenesis_muscle : np.ndarray
        Muscle total thermogenesis [W].
    q_thermogenesis_fat : np.ndarray
        Fat total thermogenesis [W].
    q_thermogenesis_skin : np.ndarray
        Skin total thermogenesis [W].
    q_skin2env_sensible : np.ndarray
        Sensible heat loss from the skin to the environment [W].
    q_skin2env_latent : np.ndarray
        Latent heat loss from the skin to the environment [W].
    q_res_sensible : np.ndarray
        Sensible heat loss by respiration [W].
    q_res_latent : np.ndarray
        Latent heat loss by respiration [W].

    """

    simulation_time: dt.timedelta | None = None
    dt: float | None = None
    t_skin_mean: float | None = None
    t_skin: JOS3BodyParts | None = None
    t_core: JOS3BodyParts | None = None
    w_mean: float | None = None
    w: JOS3BodyParts | None = None
    weight_loss_by_evap_and_res: float | None = None
    cardiac_output: float | None = None
    q_thermogenesis_total: float | None = None
    q_res: float | None = None
    q_skin2env: JOS3BodyParts | None = None
    height: float | None = None
    weight: float | None = None
    bsa: JOS3BodyParts | None = None
    fat: float | None = None
    sex: str | None = None
    age: int | None = None
    t_core_set: JOS3BodyParts | None = None
    t_skin_set: JOS3BodyParts | None = None
    t_cb: float | None = None
    t_artery: JOS3BodyParts | None = None
    t_vein: JOS3BodyParts | None = None
    t_superficial_vein: JOS3BodyParts | None = None
    t_muscle: JOS3BodyParts | None = None
    t_fat: JOS3BodyParts | None = None
    to: JOS3BodyParts | None = None
    r_t: JOS3BodyParts | None = None
    r_et: JOS3BodyParts | None = None
    tdb: JOS3BodyParts | None = None
    tr: JOS3BodyParts | None = None
    rh: JOS3BodyParts | None = None
    v: JOS3BodyParts | None = None
    par: float | None = None
    clo: JOS3BodyParts | None = None
    e_skin: JOS3BodyParts | None = None
    e_max: JOS3BodyParts | None = None
    e_sweat: JOS3BodyParts | None = None
    bf_core: JOS3BodyParts | None = None
    bf_muscle: JOS3BodyParts | None = None
    bf_fat: JOS3BodyParts | None = None
    bf_skin: JOS3BodyParts | None = None
    bf_ava_hand: float | None = None
    bf_ava_foot: float | None = None
    q_bmr_core: JOS3BodyParts | None = None
    q_bmr_muscle: JOS3BodyParts | None = None
    q_bmr_fat: JOS3BodyParts | None = None
    q_bmr_skin: JOS3BodyParts | None = None
    q_work: JOS3BodyParts | None = None
    q_shiv: JOS3BodyParts | None = None
    q_nst: JOS3BodyParts | None = None
    q_thermogenesis_core: JOS3BodyParts | None = None
    q_thermogenesis_muscle: JOS3BodyParts | None = None
    q_thermogenesis_fat: JOS3BodyParts | None = None
    q_thermogenesis_skin: JOS3BodyParts | None = None
    q_skin2env_sensible: JOS3BodyParts | None = None
    q_skin2env_latent: JOS3BodyParts | None = None
    q_res_sensible: float | None = None
    q_res_latent: float | None = None
